"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from starlette.applications import Starlette
from starlette.responses import StreamingResponse
from starlette.templating import Jinja2Templates
from starlette.staticfiles import StaticFiles
from starlette.routing import Mount
from starlette.routing import Route

from ..videogear import VideoGear
from .helper import reducer
from .helper import logger_handler
from .helper import generate_webdata
from collections import deque

import logging as log
import os, cv2, asyncio, sys


# define logger
logger = log.getLogger("WebGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class WebGear:

    """
    WebGear is a powerful ASGI Video-streamer API, that is built upon Starlette - a lightweight ASGI framework/toolkit, which is ideal 
    for building high-performance asyncio services.

    WebGear API provides a flexible but robust asyncio wrapper around Starlette ASGI application and can easily access its various components 
    independently. Thereby providing it the ability to interact with the Starlette's ecosystem of shared middleware and mountable applications 
    & seamless access to its various Response classes, Routing tables, Static Files, Templating engine(with Jinja2), etc.

    WebGear acts as robust Live Video Streaming Server that can stream live video frames to any web browser on a network in real-time. 
    It addition to this, WebGear provides a special internal wrapper around VideoGear API, which itself provides internal access to both CamGear and PiGear 
    APIs thereby granting it exclusive power for streaming frames incoming from any device/source. Also on the plus side, since WebGear has access 
    to all functions of VideoGear API, therefore it can stabilize video frames even while streaming live.


    WebGear specific parameters:

    - ** options:(dict): can be used in addition, to pass user-defined parameter to WebGear API in the form of python dictionary. Supported WebGear 
                        dictionary attributes are:

        - `custom_data_location` (str): Can be used to change/alter `default location` path to somewhere else. 

        - `overwrite_default_files` (bool): Can be used to force trigger the `Auto-generation process` to overwrite existing data-files. Remember only downloaded 
                                            files will be overwritten in this process, and any other file/folder will NOT be affected/overwritten.

        - `frame_size_reduction`: (int/float) This attribute controls the size reduction (in percentage) of the frame to be streamed on Server. 
                                                Its value can be between 0-90, and the recommended value is 40. 

        - JPEG Encoding Parameters: In WebGear, the input video frames are first encoded into Motion JPEG (M-JPEG or MJPEG) video compression format in which 
                                each video frame or interlaced field of a digital video sequence is compressed separately as a JPEG image, before sending onto 
                                a server. Therefore, WebGear API provides various attributes to have full control over JPEG encoding performance and quality, 
                                which are as follows:

            - `frame_jpeg_quality`(int) It controls the JPEG encoder quality and value varies from 0 to 100 (the higher is the better quality but 
                                            performance will be lower). Its default value is 95. 

            - `frame_jpeg_optimize`(bool) It enables various JPEG compression optimizations such as Chroma subsampling, Quantization table, etc. 
                                        Its default value is `False`. 

            - `frame_jpeg_progressive`(bool) It enables Progressive JPEG encoding instead of the Baseline. Progressive Mode. 
                                            Its default value is `False` means baseline mode. 


    VideoGear Specific parameters for WebGear:
    
        :param (boolean) enablePiCamera: set this flag to access PiGear or CamGear class respectively. 
                                        / This means the if enablePiCamera flag is `True`, PiGear class will be accessed 
                                        / and if `False`, the camGear Class will be accessed. Its default value is False.

        :param (boolean) stabilize: set this flag to enable access to VidGear's Stabilizer Class. This basically enables(if True) or disables(if False) 
                                        video stabilization in VidGear. Its default value is False.

        :param (dict) **options: can be used in addition, to pass parameter supported by VidGear's stabilizer class.
                                / Supported dict keys are: 
                                    - `SMOOTHING_RADIUS` (int) : to alter averaging window size. It handles the quality of stabilization at expense of latency and sudden panning. 
                                                            / Larger its value, less will be panning, more will be latency and vice-versa. It's default value is 30.
                                    - `BORDER_SIZE` (int) : to alter output border cropping. It's will crops the border to reduce the black borders from stabilization being too noticeable. 
                                                            / Larger its value, more will be cropping. It's default value is 0 (i.e. no cropping).          
                                    - `BORDER_TYPE` (string) : to change the border mode. Valid border types are 'black', 'reflect', 'reflect_101', 'replicate' and 'wrap'. It's default value is 'black'
    

    CamGear Specific supported parameters for WebGear:

        :param source : take the source value for CamGear Class. Its default value is 0. Valid Inputs are:
            - Index(integer): Valid index of the video device.
            - YouTube Url(string): Youtube URL as input.
            - Network_Stream_Address(string): Incoming Stream Valid Network address. 
            - GStreamer (string) videostream Support
        :param (boolean) y_tube: enables YouTube Mode in CamGear Class, i.e If enabled the class will interpret the given source string as YouTube URL. 
                                / Its default value is False.
        :param (int) backend: set the backend of the video stream (if specified). Its default value is 0.


    PiGear Specific supported parameters for WebGear:
        :param (integer) camera_num: selects the camera module index that will be used by API. 
                                /   Its default value is 0 and shouldn't be altered until unless 
                                /   if you using Raspberry Pi 3/3+ compute module in your project along with multiple camera modules. 
                                /   Furthermore, Its value can only be greater than zero, otherwise, it will throw ValueError for any negative value.
        :param (tuple) resolution: sets the resolution (width,height) in Picamera class. Its default value is (640,480).
        :param (integer) framerate: sets the framerate in Picamera class. Its default value is 25.


    Common parameters for WebGear: 
        :param (string) colorspace: set colorspace of the video stream. Its default value is None.
        :param (dict) **options: parameter supported by various API (whichever being accessed).
        :param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
        :param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
                            / This delay is essentially required for camera to warm-up. 
                            / Its default value is 0.
    """

    def __init__(
        self,
        enablePiCamera=False,
        stabilize=False,
        source=0,
        camera_num=0,
        y_tube=False,
        backend=0,
        colorspace=None,
        resolution=(640, 480),
        framerate=25,
        logging=False,
        time_delay=0,
        **options
    ):

        # initialize global params
        self.__jpeg_quality = 90  # 90% quality
        self.__jpeg_optimize = 0  # optimization off
        self.__jpeg_progressive = 0  # jpeg will be baseline instead
        self.__frame_size_reduction = 20  # 20% reduction
        self.__logging = logging

        custom_data_location = ""  # path to save data-files to custom location
        data_path = ""  # path to WebGear data-files
        overwrite_default = False

        # reformat dictionary
        options = {str(k).strip(): v for k, v in options.items()}

        # assign values to global variables if specified and valid
        if options:
            if "frame_size_reduction" in options:
                value = options["frame_size_reduction"]
                if isinstance(value, (int, float)) and value >= 0 and value <= 90:
                    self.__frame_size_reduction = value
                else:
                    logger.warning("Skipped invalid `frame_size_reduction` value!")
                del options["frame_size_reduction"]  # clean
            if "frame_jpeg_quality" in options:
                value = options["frame_jpeg_quality"]
                if isinstance(value, (int, float)) and value >= 10 and value <= 95:
                    self.__jpeg_quality = int(value)
                else:
                    logger.warning("Skipped invalid `frame_jpeg_quality` value!")
                del options["frame_jpeg_quality"]  # clean
            if "frame_jpeg_optimize" in options:
                value = options["frame_jpeg_optimize"]
                if isinstance(value, bool):
                    self.__jpeg_optimize = int(value)
                else:
                    logger.warning("Skipped invalid `frame_jpeg_optimize` value!")
                del options["frame_jpeg_optimize"]  # clean
            if "frame_jpeg_progressive" in options:
                value = options["frame_jpeg_progressive"]
                if isinstance(value, bool):
                    self.__jpeg_progressive = int(value)
                else:
                    logger.warning("Skipped invalid `frame_jpeg_progressive` value!")
                del options["frame_jpeg_progressive"]  # clean
            if "custom_data_location" in options:
                value = options["custom_data_location"]
                if isinstance(value, str):
                    assert os.access(
                        value, os.W_OK
                    ), "[WebGear:ERROR] :: Permission Denied!, cannot write WebGear data-files to '{}' directory!".format(
                        value
                    )
                    assert os.path.isdir(
                        os.path.abspath(value)
                    ), "[WebGear:ERROR] :: `custom_data_location` value must be the path to a directory and not to a file!"
                    custom_data_location = os.path.abspath(value)
                else:
                    logger.warning("Skipped invalid `custom_data_location` value!")
                del options["custom_data_location"]  # clean
            if "overwrite_default_files" in options:
                value = options["overwrite_default_files"]
                if isinstance(value, bool):
                    overwrite_default = value
                else:
                    logger.warning("Skipped invalid `overwrite_default_files` value!")
                del options["overwrite_default_files"]  # clean

        # define stream with necessary params
        self.stream = VideoGear(
            enablePiCamera=enablePiCamera,
            stabilize=stabilize,
            source=source,
            camera_num=camera_num,
            y_tube=y_tube,
            backend=backend,
            colorspace=colorspace,
            resolution=resolution,
            framerate=framerate,
            logging=logging,
            time_delay=time_delay,
            **options
        )

        # check if custom certificates path is specified
        if custom_data_location:
            data_path = generate_webdata(
                custom_data_location,
                overwrite_default=overwrite_default,
                logging=logging,
            )
        else:
            # otherwise generate suitable path
            from os.path import expanduser

            data_path = generate_webdata(
                os.path.join(expanduser("~"), ".vidgear"),
                overwrite_default=overwrite_default,
                logging=logging,
            )

        # log it
        if self.__logging:
            logger.debug(
                "`{}` is the default location for saving WebGear data-files.".format(
                    data_path
                )
            )
        if self.__logging:
            logger.debug(
                "Setting params:: Size Reduction:{}%, JPEG quality:{}%, JPEG optimizations:{}, JPEG progressive:{}".format(
                    self.__frame_size_reduction,
                    self.__jpeg_quality,
                    bool(self.__jpeg_optimize),
                    bool(self.__jpeg_progressive),
                )
            )

        # define Jinja2 templates handler
        self.__templates = Jinja2Templates(directory="{}/templates".format(data_path))

        # define custom exception handlers
        self.__exception_handlers = {404: self.__not_found, 500: self.__server_error}
        # define routing tables
        self.routes = [
            Route("/", endpoint=self.__homepage),
            Route("/video", endpoint=self.__video),
            Mount(
                "/static",
                app=StaticFiles(directory="{}/static".format(data_path)),
                name="static",
            ),
        ]
        # copying original routing tables for further validation
        self.__rt_org_copy = self.routes[:]
        # keeps check if producer loop should be running
        self.__isrunning = True

    def __call__(self):
        """
        Implements custom callable method
        """
        # validate routing tables
        assert not (self.routes is None), "Routing tables are NoneType!"
        if not isinstance(self.routes, list) or not all(
            x in self.routes for x in self.__rt_org_copy
        ):
            raise RuntimeError("Routing tables are not valid!")
        # initiate stream
        if self.__logging:
            logger.debug("Initiating Video Streaming.")
        self.stream.start()
        # return Starlette application
        if self.__logging:
            logger.debug("Running Starlette application.")
        return Starlette(
            debug=(True if self.__logging else False),
            routes=self.routes,
            exception_handlers=self.__exception_handlers,
            on_shutdown=[self.shutdown],
        )

    async def __producer(self):
        """
        Implements async frame producer.
        """
        # loop over frames
        while self.__isrunning:
            # read frame
            frame = self.stream.read()
            # break if NoneType
            if frame is None:
                break
            # reducer frames size if specified
            if self.__frame_size_reduction:
                frame = await reducer(frame, percentage=self.__frame_size_reduction)
            # handle JPEG encoding
            encodedImage = cv2.imencode(
                ".jpg",
                frame,
                [
                    cv2.IMWRITE_JPEG_QUALITY,
                    self.__jpeg_quality,
                    cv2.IMWRITE_JPEG_PROGRESSIVE,
                    self.__jpeg_progressive,
                    cv2.IMWRITE_JPEG_OPTIMIZE,
                    self.__jpeg_optimize,
                ],
            )[1].tobytes()
            # yield frame in byte format
            yield (
                b"--frame\r\nContent-Type:image/jpeg\r\n\r\n" + encodedImage + b"\r\n"
            )
            await asyncio.sleep(0.01)

    async def __video(self, scope):
        """
        Return a async video streaming response.
        """
        assert scope["type"] == "http"
        return StreamingResponse(
            self.__producer(), media_type="multipart/x-mixed-replace; boundary=frame"
        )

    async def __homepage(self, request):
        """
        Return an HTML index page.
        """
        return self.__templates.TemplateResponse("index.html", {"request": request})

    async def __not_found(self, request, exc):
        """
        Return an HTML 404 page.
        """
        return self.__templates.TemplateResponse(
            "404.html", {"request": request}, status_code=404
        )

    async def __server_error(self, request, exc):
        """
        Return an HTML 500 page.
        """
        return self.__templates.TemplateResponse(
            "500.html", {"request": request}, status_code=500
        )

    def shutdown(self):
        """
        Implements a callable to run on application shutdown
        """
        if not (self.stream is None):
            if self.__logging:
                logger.debug("Closing Video Streaming.")
            # stops producer
            self.__isrunning = False
            # stops VideoGear stream
            self.stream.stop()
            # prevent any re-iteration
            self.stream = None



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# Contains all the support functions/modules required by Vidgear

# import the necessary packages
import os, sys, requests, platform, errno
import numpy as np
from pkg_resources import parse_version
from colorlog import ColoredFormatter
from tqdm import tqdm
import logging as log

try:
    # import OpenCV Binaries
    import cv2

    # check whether OpenCV Binaries are 3.x+
    if parse_version(cv2.__version__) < parse_version("3"):
        raise ImportError(
            "[Vidgear:ERROR] :: Installed OpenCV API version(< 3.0) is not supported!"
        )
except ImportError:
    raise ImportError(
        "[Vidgear:ERROR] :: Failed to detect correct OpenCV executables, install it with `pip3 install opencv-python` command."
    )


def logger_handler():
    """
    returns logger handler
    """
    # logging formatter
    formatter = ColoredFormatter(
        "%(bold_blue)s%(name)s%(reset)s :: %(log_color)s%(levelname)s%(reset)s :: %(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            "INFO": "bold_green",
            "DEBUG": "bold_yellow",
            "WARNING": "bold_purple",
            "ERROR": "bold_red",
            "CRITICAL": "bold_red,bg_white",
        },
    )
    # define handler
    handler = log.StreamHandler()
    handler.setFormatter(formatter)
    return handler


# define logger
logger = log.getLogger("Helper")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


def check_CV_version():
    """
    returns OpenCV binary in-use version first bit 
    """
    if parse_version(cv2.__version__) >= parse_version("4"):
        return 4
    else:
        return 3


def mkdir_safe(dir, logging=False):
    """
    Simply creates directory safely
    """
    try:
        os.makedirs(dir)
        if logging:
            logger.debug("Created directory at `{}`".format(dir))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        if logging:
            logger.debug("Directory already exists at `{}`".format(dir))


def capPropId(property):
    """
    Retrieves the OpenCV property's Integer(Actual) value. 
    """
    integer_value = 0
    try:
        integer_value = getattr(cv2, property)
    except Exception as e:
        logger.exception(str(e))
        logger.critical("`{}` is not a valid OpenCV property!".format(property))
        return None
    return integer_value


def dict2Args(param_dict):
    """
    converts dict to list(args)
    """
    args = []
    for key in param_dict.keys():
        args.append(key)
        args.append(param_dict[key])
    return args


def get_valid_ffmpeg_path(
    custom_ffmpeg="", is_windows=False, ffmpeg_download_path="", logging=False
):
    """
    Validate the FFmpeg path/binaries and returns valid FFmpeg file executable location(also downloads static binaries on windows) 
    """
    final_path = ""
    if is_windows:
        # checks if current os is windows
        if custom_ffmpeg:
            # if custom FFmpeg path is given assign to local variable
            final_path += custom_ffmpeg
        else:
            # otherwise auto-download them
            try:
                if not (ffmpeg_download_path):
                    # otherwise save to Temp Directory
                    import tempfile

                    ffmpeg_download_path = tempfile.gettempdir()

                if logging:
                    logger.debug(
                        "FFmpeg Windows Download Path: {}".format(ffmpeg_download_path)
                    )

                # download Binaries
                os_bit = (
                    ("win64" if platform.machine().endswith("64") else "win32")
                    if is_windows
                    else ""
                )
                _path = download_ffmpeg_binaries(
                    path=ffmpeg_download_path, os_windows=is_windows, os_bit=os_bit
                )
                # assign to local variable
                final_path += _path

            except Exception as e:
                # log if any error occurred
                if logging:
                    logger.exception(str(e))
                    logger.debug(
                        "Error in downloading FFmpeg binaries, Check your network and Try again!"
                    )
                return False

        if os.path.isfile(final_path):
            # check if valid FFmpeg file exist
            pass
        elif os.path.isfile(os.path.join(final_path, "ffmpeg.exe")):
            # check if FFmpeg directory exists, if does, then check for valid file
            final_path = os.path.join(final_path, "ffmpeg.exe")
        else:
            # else return False
            if logging:
                logger.debug("No valid FFmpeg executables found at Custom FFmpeg path!")
            return False
    else:
        # otherwise perform test for Unix
        if custom_ffmpeg:
            # if custom FFmpeg path is given assign to local variable
            if os.path.isfile(custom_ffmpeg):
                # check if valid FFmpeg file exist
                final_path += custom_ffmpeg
            elif os.path.isfile(os.path.join(custom_ffmpeg, "ffmpeg")):
                # check if FFmpeg directory exists, if does, then check for valid file
                final_path = os.path.join(custom_ffmpeg, "ffmpeg")
            else:
                # else return False
                if logging:
                    logger.debug(
                        "No valid FFmpeg executables found at Custom FFmpeg path!"
                    )
                return False
        else:
            # otherwise assign ffmpeg binaries from system
            final_path += "ffmpeg"

    if logging:
        logger.debug("Final FFmpeg Path: {}".format(final_path))

    # Final Auto-Validation for FFmeg Binaries. returns final path if test is passed
    if validate_ffmpeg(final_path, logging=logging):
        return final_path
    else:
        return False


def download_ffmpeg_binaries(path, os_windows=False, os_bit=""):
    """
    Download and Extract FFmpeg Static Binaries for windows(if not available)
    """
    final_path = ""
    if os_windows and os_bit:
        # initialize variables
        file_url = "https://ffmpeg.zeranoe.com/builds/{}/static/ffmpeg-latest-{}-static.zip".format(
            os_bit, os_bit
        )
        file_name = os.path.join(
            os.path.abspath(path), "ffmpeg-latest-{}-static.zip".format(os_bit)
        )
        file_path = os.path.join(
            os.path.abspath(path),
            "ffmpeg-latest-{}-static/bin/ffmpeg.exe".format(os_bit),
        )
        base_path, _ = os.path.split(file_name)  # extract file base path
        # check if file already exists
        if os.path.isfile(file_path):
            final_path += file_path  # skip download if does
        else:
            # import libs
            import zipfile

            # check if given path has write access
            assert os.access(path, os.W_OK), (
                "[Helper:ERROR] :: Permission Denied, Cannot write binaries to directory = "
                + path
            )
            # remove leftovers if exists
            if os.path.isfile(file_name):
                os.remove(file_name)
            # download and write file to the given path
            with open(file_name, "wb") as f:
                logger.debug(
                    "No Custom FFmpeg path provided. Auto-Installing FFmpeg static binaries now. Please wait..."
                )
                try:
                    response = requests.get(file_url, stream=True, timeout=2)
                    response.raise_for_status()
                except Exception as e:
                    logger.exception(str(e))
                    logger.warning("Downloading Failed. Trying GitHub mirror now!")
                    file_url = "https://raw.githubusercontent.com/abhiTronix/ffmpeg-static-builds/master/windows/ffmpeg-latest-{}-static.zip".format(
                        os_bit, os_bit
                    )
                    response = requests.get(file_url, stream=True, timeout=2)
                    response.raise_for_status()
                total_length = response.headers.get("content-length")
                assert not (
                    total_length is None
                ), "[Helper:ERROR] :: Failed to retrieve files, check your Internet connectivity!"
                bar = tqdm(total=int(total_length), unit="B", unit_scale=True)
                for data in response.iter_content(chunk_size=4096):
                    f.write(data)
                    if len(data) > 0:
                        bar.update(len(data))
                bar.close()
            logger.debug("Extracting executables.")
            with zipfile.ZipFile(file_name, "r") as zip_ref:
                zip_ref.extractall(base_path)
            # perform cleaning
            os.remove(file_name)
            logger.debug("FFmpeg binaries for Windows configured successfully!")
            final_path += file_path
    # return final path
    return final_path


def validate_ffmpeg(path, logging=False):
    """
    Validate FFmeg Binaries. returns True if tests passed
    """
    try:
        # get the FFmpeg version
        version = check_output([path, "-version"])
        firstline = version.split(b"\n")[0]
        version = firstline.split(b" ")[2].strip()
        if logging:
            # log if test are passed
            logger.debug("FFmpeg validity Test Passed!")
            logger.debug(
                "Found valid FFmpeg Version: `{}` installed on this system".format(
                    version
                )
            )
    except Exception as e:
        # log if test are failed
        if logging:
            logger.exception(str(e))
            logger.warning("FFmpeg validity Test Failed!")
        return False
    return True


def check_output(*args, **kwargs):
    """
    return output from the sub-process
    """
    # silent subprocess execution
    closeNULL = 0
    import subprocess as sp

    try:
        from subprocess import DEVNULL

        closeNULL = 0
    except ImportError:
        DEVNULL = open(os.devnull, "wb")
        closeNULL = 1
    # execute command in subprocess
    process = sp.Popen(stdout=sp.PIPE, stderr=DEVNULL, *args, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    # close the process
    if closeNULL:
        DEVNULL.close()
    # if error occurred raise error
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = args[0]
        error = sp.CalledProcessError(retcode, cmd)
        error.output = output
        raise error
    return output


def generate_auth_certificates(path, overwrite=False, logging=False):

    """ 
    auto-Generates and auto-validates CURVE ZMQ keys/certificates for Netgear 
    """

    # import necessary libs
    import shutil
    import zmq.auth

    # check if path corresponds to vidgear only
    if os.path.basename(path) != ".vidgear":
        path = os.path.join(path, ".vidgear")

    # generate keys dir
    keys_dir = os.path.join(path, "keys")
    mkdir_safe(keys_dir, logging=logging)

    # generate separate public and private key dirs
    public_keys_dir = os.path.join(keys_dir, "public_keys")
    secret_keys_dir = os.path.join(keys_dir, "private_keys")

    # check if overwriting is allowed
    if overwrite:
        # delete previous certificates
        for dirs in [public_keys_dir, secret_keys_dir]:
            if os.path.exists(dirs):
                shutil.rmtree(dirs)
            mkdir_safe(dirs, logging=logging)

        # generate new keys
        server_public_file, server_secret_file = zmq.auth.create_certificates(
            keys_dir, "server"
        )
        client_public_file, client_secret_file = zmq.auth.create_certificates(
            keys_dir, "client"
        )

        # move keys to their appropriate directory respectively
        for key_file in os.listdir(keys_dir):
            if key_file.endswith(".key"):
                shutil.move(os.path.join(keys_dir, key_file), public_keys_dir)
            elif key_file.endswith(".key_secret"):
                shutil.move(os.path.join(keys_dir, key_file), secret_keys_dir)
            else:
                # clean redundant keys if present
                redundant_key = os.path.join(keys_dir, key_file)
                if os.path.isfile(redundant_key):
                    os.remove(redundant_key)
    else:
        # otherwise validate available keys
        status_public_keys = validate_auth_keys(public_keys_dir, ".key")
        status_private_keys = validate_auth_keys(secret_keys_dir, ".key_secret")

        # check if all valid keys are found
        if status_private_keys and status_public_keys:
            return (keys_dir, secret_keys_dir, public_keys_dir)

        # check if valid public keys are found
        if not (status_public_keys):
            mkdir_safe(public_keys_dir, logging=logging)

        # check if valid private keys are found
        if not (status_private_keys):
            mkdir_safe(secret_keys_dir, logging=logging)

        # generate new keys
        server_public_file, server_secret_file = zmq.auth.create_certificates(
            keys_dir, "server"
        )
        client_public_file, client_secret_file = zmq.auth.create_certificates(
            keys_dir, "client"
        )

        # move keys to their appropriate directory respectively
        for key_file in os.listdir(keys_dir):
            if key_file.endswith(".key") and not (status_public_keys):
                shutil.move(
                    os.path.join(keys_dir, key_file), os.path.join(public_keys_dir, ".")
                )
            elif key_file.endswith(".key_secret") and not (status_private_keys):
                shutil.move(
                    os.path.join(keys_dir, key_file), os.path.join(secret_keys_dir, ".")
                )
            else:
                # clean redundant keys if present
                redundant_key = os.path.join(keys_dir, key_file)
                if os.path.isfile(redundant_key):
                    os.remove(redundant_key)

    # validate newly generated keys
    status_public_keys = validate_auth_keys(public_keys_dir, ".key")
    status_private_keys = validate_auth_keys(secret_keys_dir, ".key_secret")

    # raise error is validation test fails
    if not (status_private_keys) or not (status_public_keys):
        raise RuntimeError(
            "[Helper:ERROR] :: Unable to generate valid ZMQ authentication certificates at `{}`!".format(
                keys_dir
            )
        )

    # finally return valid key paths
    return (keys_dir, secret_keys_dir, public_keys_dir)


def validate_auth_keys(path, extension):

    """
    validates and maintains ZMQ Auth Keys/Certificates
    """
    # check for valid path
    if not (os.path.exists(path)):
        return False

    # check if directory empty
    if not (os.listdir(path)):
        return False

    keys_buffer = []  # stores auth-keys

    # loop over auth-keys
    for key_file in os.listdir(path):
        key = os.path.splitext(key_file)
        # check if valid key is generated
        if key and (key[0] in ["server", "client"]) and (key[1] == extension):
            keys_buffer.append(key_file)  # store it

    # remove invalid keys if found
    if len(keys_buffer) == 1:
        os.remove(os.path.join(path, keys_buffer[0]))

    # return results
    return True if (len(keys_buffer) == 2) else False



# This is a modified algorithm/code based on findings of `Simple video stabilization using OpenCV`
# published on February 20, 2014 by nghiaho12 (http://nghiaho.com/?p=2093)

"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from collections import deque
from .helper import check_CV_version
from .helper import logger_handler
import numpy as np
import cv2
import logging as log


# define logger
logger = log.getLogger("Stabilizer")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class Stabilizer:

    """
	This is an auxiliary class that enables Real-time Video Stabilization for vidgear with minimalistic latency and at the expense of little to no 
	additional computational power requirement. The basic idea behind it is to tracks and save the salient feature array for the given number of frames 
	and then uses these anchor point to cancel out all perturbations relative to it for the incoming frames in the queue. This class relies heavily on 
	Threaded Queue mode for error-free, ultrafast frame handling and it is enabled by default.

	:param smoothing_radius` (int) : to alter averaging window size. It handles the quality of stabilization at expense of latency and sudden panning. 
									/ Larger its value, less will be panning, more will be latency and vice-versa. It's default value is 25.
	:param border_size (int) :  to create extended output border size. It's default value is `0`(no borders).			
	:param border_type (string) : to change the border mode. Valid border types are 'black', 'reflect', 'reflect_101', 'replicate' and 'wrap'. It's default value is 'black'
	:param (boolean) crop_n_zoom : to enable cropping and zooms frames to reduce the black borders from stabilization being too noticeable. 
									/It works in conjunction with the `border_size` parameter, i.e. when this parameter is enabled `border_size` 
									/ will be used for cropping border instead of making them. Its default value is False.
	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	"""

    def __init__(
        self,
        smoothing_radius=25,
        border_type="black",
        border_size=0,
        crop_n_zoom=False,
        logging=False,
    ):

        # initialize deques for handling input frames and its indexes
        self.__frame_queue = deque(maxlen=smoothing_radius)
        self.__frame_queue_indexes = deque(maxlen=smoothing_radius)

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        # define and create Adaptive histogram equalization (AHE) object for optimizations
        self.__clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))

        # initialize global vars
        self.__smoothing_radius = smoothing_radius  # averaging window, handles the quality of stabilization at expense of latency and sudden panning
        self.__smoothed_path = None  # handles the smoothed path with box filter
        self.__path = None  # handles path i.e cumulative sum of pevious_2_current transformations along a axis
        self.__transforms = []  # handles pevious_2_current transformations [dx,dy,da]
        self.__frame_transforms_smoothed = None  # handles smoothed array of pevious_2_current transformations w.r.t to frames
        self.__previous_gray = None  # handles previous gray frame
        self.__previous_keypoints = (
            None  # handles previous detect_GFTTed keypoints w.r.t previous gray frame
        )
        self.__frame_height, self.frame_width = (
            0,
            0,
        )  # handles width and height of input frames
        self.__crop_n_zoom = 0  # handles cropping and zooms frames to reduce the black borders from stabilization being too noticeable.

        # if check if crop_n_zoom defined
        if crop_n_zoom and border_size:
            self.__crop_n_zoom = border_size  # crops and zoom frame to original size
            self.__border_size = 0  # zero out border size
            self.__frame_size = None  # handles frame size for zooming
            if logging:
                logger.debug("Setting Cropping margin {} pixels".format(border_size))
        else:
            # Add output borders to frame
            self.__border_size = border_size
            if self.__logging and border_size:
                logger.debug("Setting Border size {} pixels".format(border_size))

        # define valid border modes
        border_modes = {
            "black": cv2.BORDER_CONSTANT,
            "reflect": cv2.BORDER_REFLECT,
            "reflect_101": cv2.BORDER_REFLECT_101,
            "replicate": cv2.BORDER_REPLICATE,
            "wrap": cv2.BORDER_WRAP,
        }
        # choose valid border_mode from border_type
        if border_type in ["black", "reflect", "reflect_101", "replicate", "wrap"]:
            if not crop_n_zoom:
                # initialize global border mode variable
                self.__border_mode = border_modes[border_type]
                if self.__logging and border_type != "black":
                    logger.debug("Setting Border type: {}".format(border_type))
            else:
                # log and reset to default
                if self.__logging and border_type != "black":
                    logger.debug(
                        "Setting border type is disabled if cropping is enabled!"
                    )
                self.__border_mode = border_modes["black"]
        else:
            # otherwise log if not
            if logging:
                logger.debug("Invalid input border type!")
            self.__border_mode = border_modes["black"]  # reset to default mode

        # define normalized box filter
        self.__box_filter = np.ones(smoothing_radius) / smoothing_radius

    def stabilize(self, frame):
        """
		Return stabilized video frame

		:param frame(numpy.ndarray): input video frame
		"""
        # check if frame is None
        if frame is None:
            # return if it does
            return

        # save frame size for zooming
        if self.__crop_n_zoom and self.__frame_size == None:
            self.__frame_size = frame.shape[:2]

        # initiate transformations capturing
        if not self.__frame_queue:
            # for first frame
            previous_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)  # convert to gray
            previous_gray = self.__clahe.apply(previous_gray)  # optimize gray frame
            self.__previous_keypoints = cv2.goodFeaturesToTrack(
                previous_gray,
                maxCorners=200,
                qualityLevel=0.05,
                minDistance=30.0,
                blockSize=3,
                mask=None,
                useHarrisDetector=False,
                k=0.04,
            )  # track features using GFTT
            self.__frame_height, self.frame_width = frame.shape[
                :2
            ]  # save input frame height and width
            self.__frame_queue.append(frame)  # save frame to deque
            self.__frame_queue_indexes.append(0)  # save frame index to deque
            self.__previous_gray = previous_gray[
                :
            ]  # save gray frame clone for further processing

        elif self.__frame_queue_indexes[-1] <= self.__smoothing_radius - 1:
            # for rest of frames
            self.__frame_queue.append(frame)  # save frame to deque
            self.__frame_queue_indexes.append(
                self.__frame_queue_indexes[-1] + 1
            )  # save frame index
            self.__generate_transformations()  # generate transformations
            if self.__frame_queue_indexes[-1] == self.__smoothing_radius - 1:
                # calculate smooth path once transformation capturing is completed
                for i in range(3):
                    # apply normalized box filter to the path
                    self.__smoothed_path[:, i] = self.__box_filter_convolve(
                        (self.__path[:, i]), window_size=self.__smoothing_radius
                    )
                # calculate deviation of path from smoothed path
                deviation = self.__smoothed_path - self.__path
                # save smoothed transformation
                self.__frame_transforms_smoothed = self.frame_transform + deviation
        else:
            # start applying transformations
            self.__frame_queue.append(frame)  # save frame to deque
            self.__frame_queue_indexes.append(
                self.__frame_queue_indexes[-1] + 1
            )  # save frame index
            self.__generate_transformations()  # generate transformations
            # calculate smooth path once transformation capturing is completed
            for i in range(3):
                # apply normalized box filter to the path
                self.__smoothed_path[:, i] = self.__box_filter_convolve(
                    (self.__path[:, i]), window_size=self.__smoothing_radius
                )
            # calculate deviation of path from smoothed path
            deviation = self.__smoothed_path - self.__path
            # save smoothed transformation
            self.__frame_transforms_smoothed = self.frame_transform + deviation
            # return transformation applied stabilized frame
            return self.__apply_transformations()

    def __generate_transformations(self):
        """
		Generate pevious_2_current transformations [dx,dy,da]
		"""
        frame_gray = cv2.cvtColor(
            self.__frame_queue[-1], cv2.COLOR_BGR2GRAY
        )  # retrieve current frame and convert to gray
        frame_gray = self.__clahe.apply(frame_gray)  # optimize it

        # calculate optical flow using Lucas-Kanade differential method
        curr_kps, status, error = cv2.calcOpticalFlowPyrLK(
            self.__previous_gray, frame_gray, self.__previous_keypoints, None
        )

        # select only valid keypoints
        valid_curr_kps = curr_kps[status == 1]  # current
        valid_previous_keypoints = self.__previous_keypoints[status == 1]  # previous

        # calculate optimal affine transformation between pevious_2_current keypoints
        if check_CV_version() == 3:
            # backward compatibility with OpenCV3
            transformation = cv2.estimateRigidTransform(
                valid_previous_keypoints, valid_curr_kps, False
            )
        else:
            transformation = cv2.estimateAffinePartial2D(
                valid_previous_keypoints, valid_curr_kps
            )[0]

        # check if transformation is not None
        if not (transformation is None):
            # pevious_2_current translation in x direction
            dx = transformation[0, 2]
            # pevious_2_current translation in y direction
            dy = transformation[1, 2]
            # pevious_2_current rotation in angle
            da = np.arctan2(transformation[1, 0], transformation[0, 0])
        else:
            # otherwise zero it
            dx = dy = da = 0

        # save this transformation
        self.__transforms.append([dx, dy, da])

        # calculate path from cumulative transformations sum
        self.frame_transform = np.array(self.__transforms, dtype="float32")
        self.__path = np.cumsum(self.frame_transform, axis=0)
        # create smoothed path from a copy of path
        self.__smoothed_path = np.copy(self.__path)

        # re-calculate and save GFTT keypoints for current gray frame
        self.__previous_keypoints = cv2.goodFeaturesToTrack(
            frame_gray,
            maxCorners=200,
            qualityLevel=0.05,
            minDistance=30.0,
            blockSize=3,
            mask=None,
            useHarrisDetector=False,
            k=0.04,
        )
        # save this gray frame for further processing
        self.__previous_gray = frame_gray[:]

    def __box_filter_convolve(self, path, window_size):
        """
		applies normalized linear box filter to path w.r.t to averaging window
		
		:param: path(nd.array)(cumulative sum of transformations)
		:param: averaging window size(int)
		"""
        # pad path to size of averaging window
        path_padded = np.pad(path, (window_size, window_size), "median")
        # apply linear box filter to path
        path_smoothed = np.convolve(path_padded, self.__box_filter, mode="same")
        # crop the smoothed path to original path
        path_smoothed = path_smoothed[window_size:-window_size]
        # assert if cropping is completed
        assert path.shape == path_smoothed.shape
        # return smoothed path
        return path_smoothed

    def __apply_transformations(self):
        """
		Applies affine transformation to the given frame 
		build from previously calculated transformations
		"""

        # extract frame and its index from deque
        queue_frame = self.__frame_queue.popleft()
        queue_frame_index = self.__frame_queue_indexes.popleft()

        # create border around extracted frame w.r.t border_size
        bordered_frame = cv2.copyMakeBorder(
            queue_frame,
            top=self.__border_size,
            bottom=self.__border_size,
            left=self.__border_size,
            right=self.__border_size,
            borderType=self.__border_mode,
            value=[0, 0, 0],
        )
        alpha_bordered_frame = cv2.cvtColor(
            bordered_frame, cv2.COLOR_BGR2BGRA
        )  # create alpha channel
        # extract alpha channel
        alpha_bordered_frame[:, :, 3] = 0
        alpha_bordered_frame[
            self.__border_size : self.__border_size + self.__frame_height,
            self.__border_size : self.__border_size + self.frame_width,
            3,
        ] = 255

        # extracting Transformations w.r.t frame index
        dx = self.__frame_transforms_smoothed[queue_frame_index, 0]  # x-axis
        dy = self.__frame_transforms_smoothed[queue_frame_index, 1]  # y-axis
        da = self.__frame_transforms_smoothed[queue_frame_index, 2]  # angle

        # building 2x3 transformation matrix from extracted transformations
        queue_frame_transform = np.zeros((2, 3), np.float32)
        queue_frame_transform[0, 0] = np.cos(da)
        queue_frame_transform[0, 1] = -np.sin(da)
        queue_frame_transform[1, 0] = np.sin(da)
        queue_frame_transform[1, 1] = np.cos(da)
        queue_frame_transform[0, 2] = dx
        queue_frame_transform[1, 2] = dy

        # Applying an affine transformation to the frame
        frame_wrapped = cv2.warpAffine(
            alpha_bordered_frame,
            queue_frame_transform,
            alpha_bordered_frame.shape[:2][::-1],
            borderMode=self.__border_mode,
        )

        # drop alpha channel
        frame_stabilized = frame_wrapped[:, :, :3]

        # crop and zoom
        if self.__crop_n_zoom:
            # crop stabilized frame
            frame_cropped = frame_stabilized[
                self.__crop_n_zoom : -self.__crop_n_zoom,
                self.__crop_n_zoom : -self.__crop_n_zoom,
            ]
            # zoom stabilized frame
            interpolation = (
                cv2.INTER_CUBIC if (check_CV_version() < 4) else cv2.INTER_LINEAR_EXACT
            )
            frame_stabilized = cv2.resize(
                frame_cropped, self.__frame_size[::-1], interpolation=interpolation
            )

        # finally return stabilized frame
        return frame_stabilized

    def clean(self):
        """
		clean resources(deque)
		"""
        # check if deque present
        if self.__frame_queue:
            # clear frame deque
            self.__frame_queue.clear()
            # clear frame indexes deque
            self.__frame_queue_indexes.clear()



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from .camgear import CamGear
from .helper import logger_handler
import logging as log


# define logger
logger = log.getLogger("VideoGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class VideoGear:

    """
    VideoGear API provides a special internal wrapper around VidGear's exclusive Video Stabilizer Class. Furthermore, VideoGear API can provide 
    internal access to both CamGear and PiGear APIs separated by a special flag. Thereby, this API holds the exclusive power for any incoming 
    VideoStream from any source, whether it is live or not, to stabilize it directly with minimum latency and memory requirements.


    VideoGear Specific parameters:
    
        :param (boolean) enablePiCamera: set this flag to access PiGear or CamGear API respectively. 
                                        / This means the if enablePiCamera flag is `True`, PiGear API will be accessed 
                                        / and if `False`, the camGear API will be accessed. Its default value is False.

        :param (boolean) stabilize: set this flag to enable access to VidGear's Stabilizer API. This basically enables(if True) or disables(if False) 
                                        video stabilization in VidGear. Its default value is False.

        :param (dict) **options: can be used in addition, to pass parameter supported by VidGear's stabilizer Class.
                                / Supported dict keys are: 
                                    - `SMOOTHING_RADIUS` (int) : to alter averaging window size. It handles the quality of stabilization at expense of latency and sudden panning. 
                                                            / Larger its value, less will be panning, more will be latency and vice-versa. It's default value is 30.
                                    - `BORDER_SIZE` (int) : to alter output border cropping. It's will crops the border to reduce the black borders from stabilization being too noticeable. 
                                                            / Larger its value, more will be cropping. It's default value is 0 (i.e. no cropping).          
                                    - `BORDER_TYPE` (string) : to change the border mode. Valid border types are 'black', 'reflect', 'reflect_101', 'replicate' and 'wrap'. It's default value is 'black'
        
        :param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
    

    CamGear Specific supported parameters for VideoGear:

        :param source : take the source value for CamGear API. Its default value is 0. Valid Inputs are:
            - Index(integer): Valid index of the video device.
            - YouTube Url(string): Youtube URL as input.
            - Network_Stream_Address(string): Incoming Stream Valid Network address. 
            - GStreamer (string) videostream Support
        :param (boolean) y_tube: enables YouTube Mode in CamGear API, i.e If enabled the API will interpret the given source string as YouTube URL. 
                                / Its default value is False.
        :param (int) backend: set the backend of the video stream (if specified). Its default value is 0.


    PiGear Specific supported parameters for VideoGear:
        
        :param (integer) camera_num: selects the camera module index that will be used by API. 
                                /   Its default value is 0 and shouldn't be altered until unless 
                                /   if you using Raspberry Pi 3/3+ compute module in your project along with multiple camera modules. 
                                /   Furthermore, Its value can only be greater than zero, otherwise, it will throw ValueError for any negative value.
        :param (tuple) resolution: sets the resolution (width,height) in Picamera API. Its default value is (640,480).
        :param (integer) framerate: sets the framerate in Picamera API. Its default value is 25.


    Common parameters for CamGear and PiGear: 
        :param (string) colorspace: set colorspace of the video stream. Its default value is None.
        :param (dict) **options: sets parameter supported by PiCamera or Camgear API (whichever being accessed).
        :param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
        :param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
                            / This delay is essentially required for camera to warm-up. 
                            / Its default value is 0.
    """

    def __init__(
        self,
        #VideoGear parameters
        enablePiCamera=False,
        stabilize=False,
        #PiGear parameters
        camera_num=0,
        resolution=(640, 480),
        framerate=30,
        #CamGear parameters
        source=0,
        y_tube=False,
        backend=0,
        #common parameters
        time_delay=0,
        colorspace=None,
        logging=False,
        **options
    ):

        # initialize stabilizer
        self.__stablization_mode = stabilize

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        # reformat dictionary
        options = {str(k).strip(): v for k, v in options.items()}

        if self.__stablization_mode:
            from .stabilizer import Stabilizer

            s_radius, border_size, border_type, crop_n_zoom = (
                25,
                0,
                "black",
                False,
            )  # defaults
            if options:
                if "SMOOTHING_RADIUS" in options:
                    if isinstance(options["SMOOTHING_RADIUS"], int):
                        s_radius = options[
                            "SMOOTHING_RADIUS"
                        ]  # assigsn special parameter to global variable
                    del options["SMOOTHING_RADIUS"]  # clean
                if "BORDER_SIZE" in options:
                    if isinstance(options["BORDER_SIZE"], int):
                        border_size = options[
                            "BORDER_SIZE"
                        ]  # assigsn special parameter
                    del options["BORDER_SIZE"]  # clean
                if "BORDER_TYPE" in options:
                    if isinstance(options["BORDER_TYPE"], str):
                        border_type = options[
                            "BORDER_TYPE"
                        ]  # assigsn special parameter
                    del options["BORDER_TYPE"]  # clean
                if "CROP_N_ZOOM" in options:
                    if isinstance(options["CROP_N_ZOOM"], bool):
                        crop_n_zoom = options[
                            "CROP_N_ZOOM"
                        ]  # assigsn special parameter
                    del options["CROP_N_ZOOM"]  # clean
            self.__stabilizer_obj = Stabilizer(
                smoothing_radius=s_radius,
                border_type=border_type,
                border_size=border_size,
                crop_n_zoom=crop_n_zoom,
                logging=logging,
            )
            if self.__logging:
                logger.debug(
                    "Enabling Stablization Mode for the current video source!"
                )  # log info

        if enablePiCamera:
            # only import the pigear module only if required
            from .pigear import PiGear

            # initialize the picamera stream by enabling PiGear API
            self.stream = PiGear(
                camera_num=camera_num,
                resolution=resolution,
                framerate=framerate,
                colorspace=colorspace,
                logging=logging,
                time_delay=time_delay,
                **options
            )
        else:
            # otherwise, we are using OpenCV so initialize the webcam
            # stream by activating CamGear API
            self.stream = CamGear(
                source=source,
                y_tube=y_tube,
                backend=backend,
                colorspace=colorspace,
                logging=logging,
                time_delay=time_delay,
                **options
            )

        # initialize framerate variable
        self.framerate = self.stream.framerate

    def start(self):
        # start the threaded video stream
        self.stream.start()
        return self

    def read(self):
        # return the current frame
        while self.__stablization_mode:
            frame = self.stream.read()
            if frame is None:
                break
            frame_stab = self.__stabilizer_obj.stabilize(frame)
            if not (frame_stab is None):
                return frame_stab
        return self.stream.read()

    def stop(self):
        # stop the thread and release any resources
        self.stream.stop()
        # logged
        if self.__logging:
            logger.debug("Terminating VideoGear.")
        # clean queue
        if self.__stablization_mode:
            self.__stabilizer_obj.clean()



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the packages
from threading import Thread
from pkg_resources import parse_version
import cv2, sys, time
from .helper import capPropId
from .helper import logger_handler
import logging as log


# define logger
logger = log.getLogger("PiGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class PiGear:

    """
	PiGear is similar to CamGear but exclusively made to support various Raspberry Pi Camera Modules 
	(such as OmniVision OV5647 Camera Module and Sony IMX219 Camera Module). To interface with these 
	modules correctly, PiGear provides a flexible multi-threaded wrapper around complete picamera 
	python library and provides us the ability to exploit its various features like `brightness, saturation, sensor_mode`, etc. effortlessly.

	:param (integer) camera_num: selects the camera module index that will be used by API. 
								/	Its default value is 0 and shouldn't be altered until unless 
								/	if you using Raspberry Pi 3/3+ compute module in your project along with multiple camera modules. 
								/	Furthermore, Its value can only be greater than zero, otherwise, it will throw ValueError for any negative value.
	
	:param (tuple) resolution: sets the resolution (width,height). Its default value is (640,480).

	:param (integer) framerate: sets the framerate. Its default value is 30.

	:param (string) colorspace: set colorspace of the video stream. Its default value is None.

	:param (dict) **options: sets parameter supported by PiCamera Class to the input video stream. 
							/ These attribute provides the flexibility to manipulate input raspicam video stream directly. 
							/ Parameters can be passed using this **option, allows you to pass key worded variable length of arguments to PiGear Class.

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	:param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
					/ This delay is essentially required for camera to warm-up. 
					/ Its default value is 0.

	"""

    def __init__(
        self,
        camera_num=0,
        resolution=(640, 480),
        framerate=30,
        colorspace=None,
        logging=False,
        time_delay=0,
        **options
    ):

        try:
            import picamera
            from picamera.array import PiRGBArray
            from picamera import PiCamera
        except Exception as error:
            if isinstance(error, ImportError):
                # Output expected ImportErrors.
                raise ImportError(
                    '[PiGear:ERROR] :: Failed to detect Picamera executables, install it with "pip3 install picamera" command.'
                )
            else:
                # Handle any API errors
                raise RuntimeError(
                    "[PiGear:ERROR] :: Picamera API failure: {}".format(error)
                )

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        assert (
            isinstance(framerate, (int, float)) and framerate > 5.0
        ), "[PiGear:ERROR] :: Input framerate value `{}` is a Invalid! Kindly read docs.".format(
            framerate
        )
        assert (
            isinstance(resolution, (tuple, list)) and len(resolution) == 2
        ), "[PiGear:ERROR] :: Input resolution value `{}` is a Invalid! Kindly read docs.".format(
            resolution
        )
        if not (isinstance(camera_num, int) and camera_num >= 0):
            camera_num = 0
            logger.warning(
                "Input camera_num value `{}` is invalid, Defaulting to index 0!"
            )

        # initialize the picamera stream at given index
        self.__camera = PiCamera(camera_num=camera_num)
        self.__camera.resolution = tuple(resolution)
        self.__camera.framerate = framerate
        if self.__logging:
            logger.debug(
                "Activating Pi camera at index: {} with resolution: {} & framerate: {}".format(
                    camera_num, resolution, framerate
                )
            )

        # initialize framerate variable
        self.framerate = framerate

        # initializing colorspace variable
        self.color_space = None

        # reformat dict
        options = {str(k).strip(): v for k, v in options.items()}

        # define timeout variable default value(handles hardware failures)
        self.__failure_timeout = 2.0

        # User-Defined parameter
        if options and "HWFAILURE_TIMEOUT" in options:
            # for altering timeout variable manually
            if isinstance(options["HWFAILURE_TIMEOUT"], (int, float)):
                if not (10.0 > options["HWFAILURE_TIMEOUT"] > 1.0):
                    raise ValueError(
                        "[PiGear:ERROR] :: `HWFAILURE_TIMEOUT` value can only be between 1.0 ~ 10.0"
                    )
                self.__failure_timeout = options[
                    "HWFAILURE_TIMEOUT"
                ]  # assign special parameter
                if self.__logging:
                    logger.debug(
                        "Setting HW Failure Timeout: {} seconds".format(
                            self.__failure_timeout
                        )
                    )
            del options["HWFAILURE_TIMEOUT"]  # clean

        try:
            # apply attributes to source if specified
            for key, value in options.items():
                setattr(self.__camera, key, value)
        except Exception as e:
            # Catch if any error occurred
            if self.__logging:
                logger.exception(str(e))

        # separately handle colorspace value to int conversion
        if not (colorspace is None):
            self.color_space = capPropId(colorspace.strip())
            if self.__logging and not (self.color_space is None):
                logger.debug(
                    "Enabling `{}` colorspace for this video stream!".format(
                        colorspace.strip()
                    )
                )

        # enable rgb capture array thread and capture stream
        self.__rawCapture = PiRGBArray(self.__camera, size=resolution)
        self.stream = self.__camera.capture_continuous(
            self.__rawCapture, format="bgr", use_video_port=True
        )

        # frame variable initialization
        self.frame = None
        try:
            stream = next(self.stream)
            self.frame = stream.array
            self.__rawCapture.seek(0)
            self.__rawCapture.truncate()
            # render colorspace if defined
            if not(self.frame is None) and not(self.color_space is None):
                self.frame = cv2.cvtColor(self.frame, self.color_space)
        except Exception as e:
            logger.exception(str(e))
            raise RuntimeError("[PiGear:ERROR] :: Camera Module failed to initialize!")

        # applying time delay to warm-up picamera only if specified
        if time_delay:
            time.sleep(time_delay)

        # thread initialization
        self.__thread = None

        # timer thread initialization(Keeps check on frozen thread)
        self.__timer = None
        self.__t_elasped = 0.0  # records time taken by thread

        # catching thread exceptions
        self.__exceptions = None

        # initialize termination flag
        self.__terminate = False

    def start(self):
        """
		start the thread to read frames from the video stream and initiate internal timer
		"""
        # Start frame producer thread
        self.__thread = Thread(target=self.__update, name="PiGear", args=())
        self.__thread.daemon = True
        self.__thread.start()

        # Start internal timer thread
        self.__timer = Thread(target=self.__timeit, name="PiTimer", args=())
        self.__timer.daemon = True
        self.__timer.start()

        return self

    def __timeit(self):
        """
		Keep checks on Thread excecution timing
		"""
        # assign current time
        self.__t_elasped = time.time()

        # loop until termainated
        while not (self.__terminate):
            # check for frozen thread
            if time.time() - self.__t_elasped > self.__failure_timeout:
                # log failure
                if self.__logging:
                    logger.critical("Camera Module Disconnected!")
                # prepare for clean exit
                self.__exceptions = True
                self.__terminate = True  # self-terminate

    def __update(self):
        """
		Update frames from stream
		"""
        # keep looping infinitely until the thread is terminated
        while not(self.__terminate):

            try:
                # Try to iterate next frame from generator
                stream = next(self.stream)
            except Exception:
                # catch and save any exceptions
                self.__exceptions = sys.exc_info()
                break  # exit

            # __update timer
            self.__t_elasped = time.time()

            # grab the frame from the stream and clear the stream in
            # preparation for the next frame
            frame = stream.array
            self.__rawCapture.seek(0)
            self.__rawCapture.truncate()

            # apply colorspace if specified
            if not (self.color_space is None):
                # apply colorspace to frames
                color_frame = None
                try:
                    if isinstance(self.color_space, int):
                        color_frame = cv2.cvtColor(frame, self.color_space)
                    else:
                        if self.__logging:
                            logger.warning(
                                "Global color_space parameter value `{}` is not a valid!".format(
                                    self.color_space
                                )
                            )
                        self.color_space = None
                except Exception as e:
                    # Catch if any error occurred
                    self.color_space = None
                    if self.__logging:
                        logger.exception(str(e))
                        logger.warning("Input colorspace is not a valid colorspace!")

                if not (color_frame is None):
                    self.frame = color_frame
                else:
                    self.frame = frame
            else:
                self.frame = frame

        # terminate processes
        if not (self.__terminate):
            self.__terminate = True

        # release picamera resources
        self.stream.close()
        self.__rawCapture.close()
        self.__camera.close()

    def read(self):
        """
		return the frame
		"""
        # check if there are any thread exceptions
        if not (self.__exceptions is None):
            if isinstance(self.__exceptions, bool):
                # clear frame
                self.frame = None
                # notify user about hardware failure
                raise SystemError(
                    "[PiGear:ERROR] :: Hardware failure occurred, Kindly reconnect Camera Module and restart your Pi!"
                )
            else:
                # clear frame
                self.frame = None
                # re-raise error for debugging
                error_msg = "[PiGear:ERROR] :: Camera Module API failure occured: {}".format(
                    self.__exceptions[1]
                )
                raise RuntimeError(error_msg).with_traceback(self.__exceptions[2])

        # return the frame
        return self.frame

    def stop(self):
        """
		Terminates the Read process
		"""
        if self.__logging:
            logger.debug("Terminating PiGear Processes.")

        # make sure that the threads should be terminated
        self.__terminate = True

        # stop timer thread
        if not (self.__timer is None):
            self.__timer.join()

        # handle camera thread
        if not (self.__thread is None):
            # check if hardware failure occured
            if not (self.__exceptions is None) and isinstance(self.__exceptions, bool):
                # force release picamera resources
                self.stream.close()
                self.__rawCapture.close()
                self.__camera.close()

                # properly handle thread exit
                self.__thread.join()
                self.__thread.wait()  # wait if still process is still processing some information
                self.__thread = None
            else:
                # properly handle thread exit
                self.__thread.join()



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages/libraries
from pkg_resources import parse_version
import cv2, os, sys, time
import subprocess as sp
import logging as log
from .helper import get_valid_ffmpeg_path
from .helper import capPropId
from .helper import dict2Args
from .helper import logger_handler


# define logger
logger = log.getLogger("WriteGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class WriteGear:

    """
	solely handles various powerful FFmpeg tools that allow us to do almost anything you can imagine with multimedia files. 
	With WriteGear API, you can process real-time video frames into a lossless format and specification suitable for our playback 
	in just a few lines of codes. These specifications include setting bitrate, codec, framerate, resolution, subtitles, compression, etc.  
	In addition to this, WriteGear also provides flexible access to OpenCV's VideoWriter API which provide some basic tools 
	for video frames encoding but without compression.

	WriteGear primarily operates in the following two modes:

		Compression Mode: In this mode, WriteGear utilizes FFmpeg's inbuilt encoders to encode lossless multimedia files. 
							It provides us the ability to exploit almost any available parameters available within FFmpeg, with so much ease 
							and flexibility and while doing that it robustly handles all errors/warnings quietly. 

		Non-Compression Mode: In this mode, WriteGear utilizes basic OpenCV's inbuilt VideoWriter API. Similar to compression mode, WriteGear also supports 
							all parameters manipulation available within OpenCV's VideoWriter API. But this mode lacks the ability to manipulate encoding parameters 
							and other important features like video compression, audio encoding, etc. 


	```Warning: In case, This class fails to detect valid FFmpeg executables on your system, It can automatically fallbacks to Non-Compression Mode.```


	:param output_filename (string): sets the output Video file path/filename. Valid Inputs are:

		- path: Valid path of the directory to save the output video file. Then, the class will automatically assign filename(with default extension `.mp4`).  

		- filename(with path): Valid filename(with valid extension) of the output video file. 

	```Note: Make sure to provide valid filename with path and file-extension(based on encoder(in case of FFmpeg)/fourcc(in case of OpenCV) being currently used)```

	:param compression_mode (boolean): enables/disables Video Compression Mode, i.e If enabled(means Compression Mode) WriteGear class will utilize 
						/ the FFmpeg installed/given executables to encode output video and if disabled(means Non-Compression Mode) 
						/ OpenCV's VideoWriter Class will be used. Its default value is True.

	:param custom_ffmpeg (string): (if specified) sets the custom path/directory where the `Custom FFmpeg executables` are located, works only in Compression Mode.
						/ It is compulsory to provide custom binaries on a Windows Machine, otherwise this class with automatically download and extract suitable Static FFmpeg 
						/ binaries to Temporary directory.Read VidGear Docs for more info.

	:param logging (boolean): set this flag to enable/disable error logging essential for debugging. Its default value is False.

	:param **output_params (dict): sets all properties supported by FFmpeg(in Compression Mode)/ OpenCV's VideoWriter(in Non-Compression Mode) properties 
						/ to the input video stream in CamGear Class. All the valid User-defined FFmeg/OpenCV's VideoWriter output video file parameters/properties 
						/ manipulation is currently supported. These attribute provides the flexibility to manipulate output video  directly. 
						/ In addition other special attributes to manipulate WriteGear inbuilt properties are also supported. Read VidGear Docs for more info.

	"""

    def __init__(
        self,
        output_filename="",
        compression_mode=True,
        custom_ffmpeg="",
        logging=False,
        **output_params
    ):

        # assign parameter values to class variables
        self.__compression = compression_mode
        self.__os_windows = (
            True if os.name == "nt" else False
        )  # checks if machine in-use is running windows os or not

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        # initialize various important class variables
        self.__output_parameters = {}
        self.__inputheight = None
        self.__inputwidth = None
        self.__inputchannels = None
        self.__inputframerate = 0
        self.__output_dimensions = None
        self.__process = None  # handle process to be frames written
        self.__DEVNULL = (
            None  # handles silent execution of FFmpeg (if logging is disabled)
        )
        self.__cmd = ""  # handle FFmpeg Pipe command
        self.__ffmpeg = ""  # handle valid FFmpeg binaries location
        self.__initiate = (
            True  # initiate one time process for valid process initialization
        )

        # handles output file name (if not given)
        if not output_filename:
            raise ValueError(
                "[WriteGear:ERROR] :: Kindly provide a valid `output_filename` value. Refer Docs for more information."
            )
        elif output_filename and os.path.isdir(
            output_filename
        ):  # check if directory path is given instead
            output_filename = os.path.join(
                output_filename, "VidGear-{}.mp4".format(time.strftime("%Y%m%d-%H%M%S"))
            )  # auto-assign valid name and adds it to path
        else:
            pass

        # some definitions and assigning output file absolute path to class variable
        _filename = os.path.abspath(output_filename)
        self.__out_file = _filename
        basepath, _ = os.path.split(
            _filename
        )  # extract file base path for debugging ahead

        if output_params:
            # handle user defined output dimensions(must be a tuple or list)
            if output_params and "-output_dimensions" in output_params:
                self.__output_dimensions = output_params[
                    "-output_dimensions"
                ]  # assign special parameter to global variable
                del output_params["-output_dimensions"]  # clean

            # cleans and reformat output parameters
            self.__output_parameters = {
                str(k).strip().lower(): str(v).strip() for k, v in output_params.items()
            }

        # handles FFmpeg binaries validity tests
        if self.__compression:

            if self.__logging:
                logger.debug(
                    "Compression Mode is enabled therefore checking for valid FFmpeg executables."
                )
                logger.debug(self.__output_parameters)

            # handles where to save the downloaded FFmpeg Static Binaries on Windows(if specified)
            ffmpeg_download_path_ = ""
            if (
                self.__output_parameters
                and "-ffmpeg_download_path" in self.__output_parameters
            ):
                ffmpeg_download_path_ += self.__output_parameters[
                    "-ffmpeg_download_path"
                ]
                del self.__output_parameters["-ffmpeg_download_path"]  # clean

            # handle input framerate if specified
            if (
                self.__output_parameters
                and "-input_framerate" in self.__output_parameters
            ):
                self.__inputframerate = float(
                    self.__output_parameters["-input_framerate"]
                )
                del self.__output_parameters["-input_framerate"]  # clean

            # validate the FFmpeg path/binaries and returns valid FFmpeg file executable location(also downloads static binaries on windows)
            actual_command = get_valid_ffmpeg_path(
                custom_ffmpeg,
                self.__os_windows,
                ffmpeg_download_path=ffmpeg_download_path_,
                logging=self.__logging,
            )

            # check if valid path returned
            if actual_command:
                self.__ffmpeg += actual_command  # assign it to class variable
                if self.__logging:
                    logger.debug(
                        "Found valid FFmpeg executables: `{}`.".format(self.__ffmpeg)
                    )
            else:
                # otherwise disable Compression Mode
                logger.warning(
                    "Disabling Compression Mode since no valid FFmpeg executables found on this machine!"
                )
                if self.__logging and not self.__os_windows:
                    logger.debug(
                        "Kindly install working FFmpeg or provide a valid custom FFmpeg binary path. See docs for more info."
                    )
                self.__compression = False  # compression mode disabled

        # validate this class has the access rights to specified directory or not
        assert os.access(basepath, os.W_OK), (
            "[WriteGear:ERROR] :: Permission Denied: Cannot write to directory: "
            + basepath
        )

        # display confirmation if logging is enabled/disabled
        if self.__compression and self.__ffmpeg:
            self.__DEVNULL = open(os.devnull, "wb")
            if self.__logging:
                logger.debug("Compression Mode is configured properly!")
        else:
            if self.__logging:
                logger.debug(
                    "Compression Mode is disabled, Activating OpenCV built-in Writer!"
                )

    def write(self, frame, rgb_mode=False):

        """  
		pipelines ndarray frames to valid Video File. Utilizes FFmpeg in Compression Mode and OpenCV VideoWriter Class in Non-Compression Mode

		:param frame (ndarray): valid frame
		:param rgb_mode (boolean): set this flag to enable rgb_mode, i.e. specifies that incoming frames are of RGB format(instead of default BGR). Its default value is False.

		"""
        if frame is None:  # NoneType frames will be skipped
            return

        # get height, width and number of channels of current frame
        height, width = frame.shape[:2]
        channels = frame.shape[-1] if frame.ndim == 3 else 1

        # assign values to class variables on first run
        if self.__initiate:
            self.__inputheight = height
            self.__inputwidth = width
            self.__inputchannels = channels
            if self.__logging:
                logger.debug(
                    "InputFrame => Height:{} Width:{} Channels:{}".format(
                        self.__inputheight, self.__inputwidth, self.__inputchannels
                    )
                )

        # validate size of frame
        if height != self.__inputheight or width != self.__inputwidth:
            raise ValueError("[WriteGear:ERROR] :: All frames must have same size!")
        # validate number of channels
        if channels != self.__inputchannels:
            raise ValueError(
                "[WriteGear:ERROR] :: All frames must have same number of channels!"
            )

        if self.__compression:
            # checks if compression mode is enabled

            # initiate FFmpeg process on first run
            if self.__initiate:
                # start pre-processing and initiate process
                self.__Preprocess(channels, rgb=rgb_mode)
                # Check status of the process
                assert self.__process is not None

            # write the frame
            try:
                self.__process.stdin.write(frame.tostring())
            except (OSError, IOError):
                # log something is wrong!
                logger.error(
                    "BrokenPipeError caught, Wrong values passed to FFmpeg Pipe, Kindly Refer Docs!"
                )
                self.__DEVNULL.close()
                raise ValueError  # for testing purpose only
        else:
            # otherwise initiate OpenCV's VideoWriter Class
            if self.__initiate:
                # start VideoWriter Class process
                self.__startCV_Process()
                # Check status of the process
                assert self.__process is not None
                if self.__logging:
                    # log OpenCV warning
                    logger.info(
                        "RGBA and 16-bit grayscale video frames are not supported by OpenCV yet, switch to `compression_mode` to use them!"
                    )
            # write the frame
            self.__process.write(frame)

    def __Preprocess(self, channels, rgb=False):
        """
		pre-processing FFmpeg parameters
		
		:param channels (int): Number of channels
		:param rgb_mode (boolean): set this flag to enable rgb_mode, Its default value is False.
		"""
        # turn off initiate flag
        self.__initiate = False
        # initialize input parameters
        input_parameters = {}

        # handle dimensions
        dimensions = ""
        if self.__output_dimensions is None:  # check if dimensions are given
            dimensions += "{}x{}".format(
                self.__inputwidth, self.__inputheight
            )  # auto derive from frame
        else:
            dimensions += "{}x{}".format(
                self.__output_dimensions[0], self.__output_dimensions[1]
            )  # apply if defined
        input_parameters["-s"] = str(dimensions)

        # handles pix_fmt based on channels(HACK)
        if channels == 1:
            input_parameters["-pix_fmt"] = "gray"
        elif channels == 2:
            input_parameters["-pix_fmt"] = "ya8"
        elif channels == 3:
            input_parameters["-pix_fmt"] = "rgb24" if rgb else "bgr24"
        elif channels == 4:
            input_parameters["-pix_fmt"] = "rgba" if rgb else "bgra"
        else:
            raise ValueError(
                "[WriteGear:ERROR] :: Frames with channels outside range 1-to-4 are not supported!"
            )

        if self.__inputframerate > 5:
            # set input framerate - minimum threshold is 5.0
            if self.__logging:
                logger.debug(
                    "Setting Input FrameRate = {}".format(self.__inputframerate)
                )
            input_parameters["-framerate"] = str(self.__inputframerate)

        # initiate FFmpeg process
        self.__startFFmpeg_Process(
            input_params=input_parameters, output_params=self.__output_parameters
        )

    def __startFFmpeg_Process(self, input_params, output_params):

        """
		Start FFmpeg process

		:param input_params (dict): Input parameters
		:param output_params (dict): Output parameters
		"""

        # convert input parameters to list
        input_parameters = dict2Args(input_params)

        # pre-assign default encoder parameters (if not assigned by user).
        if "-vcodec" not in output_params:
            output_params["-vcodec"] = "libx264"
        if output_params["-vcodec"] in ["libx264", "libx265"]:
            if "-crf" in output_params:
                pass
            else:
                output_params["-crf"] = "18"
            if "-preset" in output_params:
                pass
            else:
                output_params["-preset"] = "fast"
        # convert output parameters to list
        output_parameters = dict2Args(output_params)
        # format command
        cmd = (
            [self.__ffmpeg, "-y"]
            + ["-f", "rawvideo", "-vcodec", "rawvideo"]
            + input_parameters
            + ["-i", "-"]
            + output_parameters
            + [self.__out_file]
        )
        # assign value to class variable
        self.__cmd += " ".join(cmd)
        # Launch the FFmpeg process
        if self.__logging:
            logger.debug("Executing FFmpeg command: `{}`".format(self.__cmd))
            # In debugging mode
            self.__process = sp.Popen(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=None)
        else:
            # In silent mode
            self.__process = sp.Popen(
                cmd, stdin=sp.PIPE, stdout=self.__DEVNULL, stderr=sp.STDOUT
            )

    def execute_ffmpeg_cmd(self, cmd=None):
        """
		Executes custom FFmpeg process

		:param cmd(list): custom command with input as list  
		"""
        # check if valid command
        if cmd is None:
            logger.warning("Input FFmpeg command is empty, Nothing to execute!")
            return
        else:
            if not (isinstance(cmd, list)):
                raise ValueError(
                    "[WriteGear:ERROR] :: Invalid input FFmpeg command datatype! Kindly read docs."
                )

        # check if Compression Mode is enabled
        if not (self.__compression):
            raise RuntimeError(
                "[WriteGear:ERROR] :: Compression Mode is disabled, Kindly enable it to access this function!"
            )

        # add configured FFmpeg path
        cmd = [self.__ffmpeg] + cmd

        try:
            # write to pipeline
            if self.__logging:
                logger.debug("Executing FFmpeg command: `{}`".format(" ".join(cmd)))
                # In debugging mode
                sp.call(cmd, stdin=sp.PIPE, stdout=sp.PIPE, stderr=None)
            else:
                sp.call(cmd, stdin=sp.PIPE, stdout=self.__DEVNULL, stderr=sp.STDOUT)
        except (OSError, IOError):
            # log something is wrong!
            logger.error(
                "BrokenPipeError caught, Wrong command passed to FFmpeg Pipe, Kindly Refer Docs!"
            )
            self.__DEVNULL.close()
            raise ValueError  # for testing purpose only

    def __startCV_Process(self):
        """
		Start OpenCV VideoWriter Class process
		"""

        # turn off initiate flag
        self.__initiate = False

        # initialize essential parameter variables
        FPS = 0
        BACKEND = ""
        FOURCC = 0
        COLOR = True

        # pre-assign default encoder parameters (if not assigned by user).
        if "-fourcc" not in self.__output_parameters:
            FOURCC = cv2.VideoWriter_fourcc(*"MJPG")
        if "-fps" not in self.__output_parameters:
            FPS = 25

        # auto assign dimensions
        HEIGHT = self.__inputheight
        WIDTH = self.__inputwidth

        # assign parameter dict values to variables
        try:
            for key, value in self.__output_parameters.items():
                if key == "-fourcc":
                    FOURCC = cv2.VideoWriter_fourcc(*(value.upper()))
                elif key == "-fps":
                    FPS = int(value)
                elif key == "-backend":
                    BACKEND = capPropId(value.upper())
                elif key == "-color":
                    COLOR = bool(value)
                else:
                    pass

        except Exception as e:
            # log if something is wrong
            if self.__logging:
                logger.exception(str(e))
            raise ValueError(
                "[WriteGear:ERROR] :: Wrong Values passed to OpenCV Writer, Kindly Refer Docs!"
            )

        if self.__logging:
            # log values for debugging
            logger.debug(
                "FILE_PATH: {}, FOURCC = {}, FPS = {}, WIDTH = {}, HEIGHT = {}, BACKEND = {}".format(
                    self.__out_file, FOURCC, FPS, WIDTH, HEIGHT, BACKEND
                )
            )

        # start different process for with/without Backend.
        if BACKEND:
            self.__process = cv2.VideoWriter(
                self.__out_file,
                apiPreference=BACKEND,
                fourcc=FOURCC,
                fps=FPS,
                frameSize=(WIDTH, HEIGHT),
                isColor=COLOR,
            )
        else:
            self.__process = cv2.VideoWriter(
                self.__out_file,
                fourcc=FOURCC,
                fps=FPS,
                frameSize=(WIDTH, HEIGHT),
                isColor=COLOR,
            )

    def close(self):
        """
		Terminates the Write process
		"""
        if self.__logging:
            logger.debug("Terminating WriteGear Processes.")

        if self.__compression:
            # if Compression Mode is enabled
            if self.__process is None:
                return  # no process was initiated at first place
            if self.__process.poll() is not None:
                return  # process was already dead
            if self.__process.stdin:
                self.__process.stdin.close()  # close `stdin` output
            if self.__output_parameters and "-i" in self.__output_parameters:
                self.__process.terminate()
                self.__process.wait()  # wait if still process is still processing some information
                self.__process = None
                self.__DEVNULL.close()  # close it
            else:
                self.__process.wait()  # wait if still process is still processing some information
                self.__process = None
                self.__DEVNULL.close()  # close it
        else:
            # if Compression Mode is disabled
            if self.__process is None:
                return  # no process was initiated at first place
            self.__process.release()  # close it



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from threading import Thread
from pkg_resources import parse_version
from .helper import generate_auth_certificates
from .helper import logger_handler
from collections import deque
import numpy as np
import cv2, time, os, random
import logging as log


# define logger
logger = log.getLogger("NetGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class NetGear:

    """
    NetGear is exclusively designed to transfer video frames synchronously between interconnecting systems over the network in real-time. 
    This is achieved by implementing a high-level wrapper around PyZmQ python library that contains python bindings for ZeroMQ - a 
    high-performance asynchronous distributed messaging library that aim to be used in distributed or concurrent applications. 
    It provides a message queue, but unlike message-oriented middleware, a ZeroMQ system can run without a dedicated message broker. 
    Furthermore, NetGear currently supports three ZeroMQ messaging patterns: i.e zmq.PAIR, zmq.REQ and zmq.REP,and zmq.PUB,zmq.SUB whereas
    supported protocol are: 'tcp', 'upd', 'pgm', 'inproc', 'ipc'.

    Multi-Server Mode:  This mode enables NetGear API to robustly handle multiple servers at once through exclusive Publish/Subscribe (zmq.PUB/zmq.SUB) 
                        messaging pattern for seamless Live Streaming across various device at the same time. Each device new server on network is 
                        identified using its unique port address. Also, when all the connected servers on the network get disconnected, the client 
                        itself automatically exits too. This mode can be activated through`multiserver_mode` option boolean attribute during 
                        NetGear API initialization easily.

    Secure Mode: This mode provides secure ZeroMQ's Security Layers for NetGear API that enables strong encryption on data, and (as far as we know) unbreakable 
                 authentication between the Server and the Client with the help of custom auth certificates/keys. It's default value is `Grassland`:0 => which means no
                 security at all. 

                 This mode supports the two most powerful ZMQ security mechanisms:

                 * `StoneHouse`: 1 => which switches to the "CURVE" security mechanism, giving us strong encryption on data, and unbreakable authentication. 
                                      Stonehouse is the minimum you would use over public networks and assures clients that they are speaking to an authentic server while allowing any client 
                                      to connect. It is less secure but at the same time faster than IronHouse security mechanism.

                 * `IronHouse`: 2 => which extends Stonehouse with client public key authentication. This is the strongest security model ZMQ have today, 
                                        protecting against every attack we know about, except end-point attacks (where an attacker plants spyware on a machine 
                                        to capture data before it's encrypted, or after it's decrypted). IronHouse enhanced security comes at a price of additional latency.


    :param address(string): sets the valid network address of the server/client. Network addresses unique identifiers across the 
                            network. Its default value of this parameter is based on mode it is working, 'localhost' for Send Mode
                            and `*` for Receive Mode.

    :param port(string/dict/list): sets the valid network port of the server/client. A network port is a number that identifies one side 
                            of a connection between two devices on network. It is used determine to which process or application 
                            a message should be delivered. In Multi-Server Mode a unique port number must required at each server, and a
                            list/tuple of port addresses of each connected server is required at clients end.

    :param protocol(string): sets the valid messaging protocol between server and client. A network protocol is a set of established rules
                             that dictates how to format, transmit and receive data so computer network devices - from servers and 
                             routers to endpoints - can communicate regardless of the differences in their underlying infrastructures, 
                             designs or standards. Supported protocol are: 'tcp', 'upd', 'pgm', 'inproc', 'ipc'. Its default value is `tcp`.

    :param pattern(int): sets the supported messaging pattern(flow of communication) between server and client. Messaging patterns are network 
                            oriented architectural pattern that describes the flow of communication between interconnecting systems.
                            vidgear provides access to ZeroMQ's pre-optimized sockets which enables you to take advantage of these patterns.
                            The supported patterns are:
                                0: zmq.PAIR -> In this, the communication is bidirectional. There is no specific state stored within the socket. 
                                                There can only be one connected peer.The server listens on a certain port and a client connects to it.
                                1. zmq.REQ/zmq.REP -> In this, ZMQ REQ sockets can connect to many servers. The requests will be
                                                        interleaved or distributed to both the servers. socket zmq.REQ will block 
                                                        on send unless it has successfully received a reply back and socket zmq.REP 
                                                        will block on recv unless it has received a request.
                                2. zmq.PUB,zmq.SUB -> Publish/Subscribe is another classic pattern where senders of messages, called publishers, 
                                                        do not program the messages to be sent directly to specific receivers, called subscribers. 
                                                        Messages are published without the knowledge of what or if any subscriber of that knowledge exists.
                            Its default value is `0`(i.e zmq.PAIR).

    :param (boolean) receive_mode: set this flag to select the Netgear's Mode of operation. This basically activates `Receive Mode`(if True) and `Send Mode`(if False). 
                                    Furthermore `recv()` function will only works when this flag is enabled(i.e. `Receive Mode`) and `send()` function will only works 
                                    when this flag is disabled(i.e.`Send Mode`). Checkout VidGear docs for usage details.
                                    Its default value is False(i.e. Send Mode is activated by default).

    :param **options(dict): can be used to pass flexible parameters to NetGear API. 
                            This attribute provides the flexibility to manipulate ZeroMQ internal parameters 
                            directly. Checkout vidgear docs for usage details.

    :param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

    """

    def __init__(
        self,
        address=None,
        port=None,
        protocol=None,
        pattern=0,
        receive_mode=False,
        logging=False,
        **options
    ):

        try:
            # import PyZMQ library
            import zmq

            # import ZMQError
            from zmq.error import ZMQError

            # assign values to global variable for further use
            self.__zmq = zmq
            self.__ZMQError = ZMQError
        except ImportError as error:
            # raise error
            raise ImportError(
                "[NetGear:ERROR] :: pyzmq python library not installed. Kindly install it with `pip install pyzmq` command."
            )

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        # define valid messaging patterns => `0`: zmq.PAIR, `1`:(zmq.REQ,zmq.REP), and `1`:(zmq.SUB,zmq.PUB)
        valid_messaging_patterns = {
            0: (zmq.PAIR, zmq.PAIR),
            1: (zmq.REQ, zmq.REP),
            2: (zmq.PUB, zmq.SUB),
        }

        # initialize messaging pattern
        msg_pattern = None
        # check whether user-defined messaging pattern is valid
        if isinstance(pattern, int) and pattern in valid_messaging_patterns:
            # assign value
            msg_pattern = valid_messaging_patterns[pattern]
            self.__pattern = pattern  # add it to global variable for further use
        else:
            # otherwise default to 0:`zmq.PAIR`
            self.__pattern = 0
            msg_pattern = valid_messaging_patterns[self.__pattern]
            # log it
            if self.__logging:
                logger.warning(
                    "Wrong pattern value, Defaulting to `zmq.PAIR`! Kindly refer Docs for more Information."
                )

        # check  whether user-defined messaging protocol is valid
        if not (protocol in ["tcp", "ipc"]):
            # else default to `tcp` protocol
            protocol = "tcp"
            # log it
            if self.__logging:
                logger.warning(
                    "Protocol is invalid or not provided. Defaulting to `tcp` protocol!"
                )

        # generate random device id
        self.__id = "".join(random.choice("0123456789ABCDEF") for i in range(5))

        self.__msg_flag = 0  # handles connection flags
        self.__msg_copy = False  # handles whether to copy data
        self.__msg_track = False  # handles whether to track packets

        self.__multiserver_mode = False  # handles multiserver_mode state
        recv_filter = ""  # user-defined filter to allow specific port/servers only in multiserver_mode

        # define bi-directional data transmission mode
        self.__bi_mode = False  # handles bi_mode state
        self.__bi_data = None  # handles return data

        # define valid  ZMQ security mechanisms => `0`: Grasslands, `1`:StoneHouse, and `1`:IronHouse
        valid_security_mech = {0: "Grasslands", 1: "StoneHouse", 2: "IronHouse"}
        self.__secure_mode = 0  # handles ZMQ security layer status
        auth_cert_dir = ""  # handles valid ZMQ certificates dir
        self.__auth_publickeys_dir = ""  # handles valid ZMQ public certificates dir
        self.__auth_secretkeys_dir = ""  # handles valid ZMQ private certificates dir
        overwrite_cert = False  # checks if certificates overwriting allowed
        custom_cert_location = ""  # handles custom ZMQ certificates path

        # handle force socket termination if there's latency in network
        self.__force_close = False

        # define stream compression handlers
        self.__compression = ""  # disabled by default
        self.__compression_params = None

        # reformat dict
        options = {str(k).strip(): v for k, v in options.items()}

        # assign values to global variables if specified and valid
        for key, value in options.items():
            if key == "multiserver_mode" and isinstance(value, bool):
                if pattern > 0:
                    # multi-server mode
                    self.__multiserver_mode = value
                else:
                    self.__multiserver_mode = False
                    logger.critical("Multi-Server is disabled!")
                    raise ValueError(
                        "[NetGear:ERROR] :: `{}` pattern is not valid when Multi-Server Mode is enabled. Kindly refer Docs for more Information.".format(
                            pattern
                        )
                    )
            elif key == "filter" and isinstance(value, str):
                # custom filter in multi-server mode
                recv_filter = value

            elif (
                key == "secure_mode"
                and isinstance(value, int)
                and (value in valid_security_mech)
            ):
                # secure mode
                assert zmq.zmq_version_info() >= (
                    4,
                    0,
                ), "[NetGear:ERROR] :: ZMQ Security feature is not supported in libzmq version < 4.0."
                self.__secure_mode = value
            elif key == "custom_cert_location" and isinstance(value, str):
                # custom auth certificates path
                assert os.access(
                    value, os.W_OK
                ), "[NetGear:ERROR] :: Permission Denied!, cannot write ZMQ authentication certificates to '{}' directory!".format(
                    value
                )
                assert os.path.isdir(
                    os.path.abspath(value)
                ), "[NetGear:ERROR] :: `custom_cert_location` value must be the path to a valid directory!"
                custom_cert_location = os.path.abspath(value)
            elif key == "overwrite_cert" and isinstance(value, bool):
                # enable/disable auth certificate overwriting
                overwrite_cert = value

            # handle encoding and decoding if specified
            elif (
                key == "compression_format"
                and isinstance(value, str)
                and value.lower().strip() in [".jpg", ".jpeg", ".bmp", ".png"]
            ):  # few are supported
                # enable encoding
                if not (receive_mode):
                    self.__compression = value.lower().strip()
            elif key == "compression_param":
                # specify encoding/decoding params
                if receive_mode and isinstance(value, int):
                    self.__compression_params = value
                    if self.__logging:
                        logger.debug("Decoding flag: {}.".format(value))
                elif not (receive_mode) and isinstance(value, (list, tuple)):
                    if self.__logging:
                        logger.debug("Encoding parameters: {}.".format(value))
                    self.__compression_params = list(value)
                else:
                    if self.__logging:
                        logger.warning(
                            "Invalid compression parameters: {} skipped!".format(value)
                        )
                    self.__compression_params = (
                        cv2.IMREAD_COLOR if receive_mode else []
                    )  # skip to defaults

            # enable bi-directional data transmission if specified
            elif key == "bidirectional_mode" and isinstance(value, bool):
                # check if pattern is valid
                if pattern < 2:
                    self.__bi_mode = True
                else:
                    self.__bi_mode = False
                    logger.critical("Bi-Directional data transmission is disabled!")
                    raise ValueError(
                        "[NetGear:ERROR] :: `{}` pattern is not valid when Bi-Directional Mode is enabled. Kindly refer Docs for more Information.".format(
                            pattern
                        )
                    )

            # enable force socket closing if specified
            elif key == "force_terminate" and isinstance(value, bool):
                # check if address is local
                if address is None:
                    self.__force_close = False
                    logger.critical("Force termination is disabled for local servers!")
                else:
                    self.__force_close = True
                    if self.__logging:
                        logger.warning(
                            "Force termination is enabled for this connection!"
                        )

            # various ZMQ flags
            elif key == "flag" and isinstance(value, int):
                self.__msg_flag = value
            elif key == "copy" and isinstance(value, bool):
                self.__msg_copy = value
            elif key == "track" and isinstance(value, bool):
                self.__msg_track = value
            else:
                pass

        # handle secure mode
        if self.__secure_mode:
            # import important libs
            import zmq.auth
            from zmq.auth.thread import ThreadAuthenticator

            # log if overwriting is enabled
            if overwrite_cert:
                if not receive_mode:
                    if self.__logging:
                        logger.warning(
                            "Overwriting ZMQ Authentication certificates over previous ones!"
                        )
                else:
                    overwrite_cert = False
                    if self.__logging:
                        logger.critical(
                            "Overwriting ZMQ Authentication certificates is disabled for Client-end!"
                        )

            # generate and validate certificates path
            try:
                # check if custom certificates path is specified
                if custom_cert_location:
                    (
                        auth_cert_dir,
                        self.__auth_secretkeys_dir,
                        self.__auth_publickeys_dir,
                    ) = generate_auth_certificates(
                        custom_cert_location, overwrite=overwrite_cert, logging=logging
                    )
                else:
                    # otherwise auto-generate suitable path
                    from os.path import expanduser

                    (
                        auth_cert_dir,
                        self.__auth_secretkeys_dir,
                        self.__auth_publickeys_dir,
                    ) = generate_auth_certificates(
                        os.path.join(expanduser("~"), ".vidgear"),
                        overwrite=overwrite_cert,
                        logging=logging,
                    )

                # log it
                if self.__logging:
                    logger.debug(
                        "`{}` is the default location for storing ZMQ authentication certificates/keys.".format(
                            auth_cert_dir
                        )
                    )
            except Exception as e:
                # catch if any error occurred
                logger.exception(str(e))
                # also disable secure mode
                self.__secure_mode = 0
                logger.critical(
                    "ZMQ Security Mechanism is disabled for this connection!"
                )
        else:
            # log if disabled
            if self.__logging:
                logger.warning(
                    "ZMQ Security Mechanism is disabled for this connection!"
                )

        # handle bi_mode
        if self.__bi_mode:
            # disable bi_mode if multi-server is enabled
            if self.__multiserver_mode:
                self.__bi_mode = False
                logger.critical(
                    "Bi-Directional Data Transmission is disabled when Multi-Server Mode is Enabled due to incompatibility!"
                )
            else:
                # enable force termination by default
                self.__force_close = True
                if self.__logging:
                    logger.warning(
                        "Force termination is enabled for this connection by default!"
                    )
                    logger.debug(
                        "Bi-Directional Data Transmission is enabled for this connection!"
                    )

        # initialize termination flag
        self.__terminate = False

        # initialize exit_loop flag
        self.__exit_loop = False

        # initialize and assign receive mode to global variable
        self.__receive_mode = receive_mode

        # define messaging context instance
        self.__msg_context = zmq.Context.instance()

        # check whether `receive_mode` is enabled by user
        if receive_mode:

            # if does than define connection address
            if address is None:
                address = "*"  # define address

            # check if multiserver_mode is enabled
            if self.__multiserver_mode:
                # check if unique server port address list/tuple is assigned or not in multiserver_mode
                if port is None or not isinstance(port, (tuple, list)):
                    # raise error if not
                    raise ValueError(
                        "[NetGear:ERROR] :: Incorrect port value! Kindly provide a list/tuple of ports while Multi-Server mode is enabled. For more information refer VidGear docs."
                    )
                else:
                    # otherwise log it
                    logger.debug(
                        "Enabling Multi-Server Mode at PORTS: {}!".format(port)
                    )
                # create port address buffer for keeping track of incoming server's port
                self.__port_buffer = []
            else:
                # otherwise assign local port address if None
                if port is None:
                    port = "5555"

            try:
                # initiate and handle secure mode
                if self.__secure_mode > 0:
                    # start an authenticator for this context
                    auth = ThreadAuthenticator(self.__msg_context)
                    auth.start()
                    auth.allow(str(address))  # allow current address

                    # check if `IronHouse` is activated
                    if self.__secure_mode == 2:
                        # tell authenticator to use the certificate from given valid dir
                        auth.configure_curve(
                            domain="*", location=self.__auth_publickeys_dir
                        )
                    else:
                        # otherwise tell the authenticator how to handle the CURVE requests, if `StoneHouse` is activated
                        auth.configure_curve(
                            domain="*", location=zmq.auth.CURVE_ALLOW_ANY
                        )

                # initialize and define thread-safe messaging socket
                self.__msg_socket = self.__msg_context.socket(msg_pattern[1])
                if self.__pattern == 2 and not (self.__secure_mode):
                    self.__msg_socket.set_hwm(1)

                if self.__multiserver_mode:
                    # if multiserver_mode is enabled, then assign port addresses to zmq socket
                    for pt in port:
                        # enable specified secure mode for the zmq socket
                        if self.__secure_mode > 0:
                            # load server key
                            server_secret_file = os.path.join(
                                self.__auth_secretkeys_dir, "server.key_secret"
                            )
                            server_public, server_secret = zmq.auth.load_certificate(
                                server_secret_file
                            )
                            # load  all CURVE keys
                            self.__msg_socket.curve_secretkey = server_secret
                            self.__msg_socket.curve_publickey = server_public
                            # enable CURVE connection for this socket
                            self.__msg_socket.curve_server = True

                        # define socket options
                        if self.__pattern == 2:
                            self.__msg_socket.setsockopt_string(
                                zmq.SUBSCRIBE, recv_filter
                            )

                        # bind socket to given server protocol, address and ports
                        self.__msg_socket.bind(
                            protocol + "://" + str(address) + ":" + str(pt)
                        )

                    # define socket optimizer
                    self.__msg_socket.setsockopt(zmq.LINGER, 0)

                else:
                    # enable specified secure mode for the zmq socket
                    if self.__secure_mode > 0:
                        # load server key
                        server_secret_file = os.path.join(
                            self.__auth_secretkeys_dir, "server.key_secret"
                        )
                        server_public, server_secret = zmq.auth.load_certificate(
                            server_secret_file
                        )
                        # load  all CURVE keys
                        self.__msg_socket.curve_secretkey = server_secret
                        self.__msg_socket.curve_publickey = server_public
                        # enable CURVE connection for this socket
                        self.__msg_socket.curve_server = True

                    # define exclusive socket options for patterns
                    if self.__pattern == 2:
                        self.__msg_socket.setsockopt_string(zmq.SUBSCRIBE, "")

                    # bind socket to given protocol, address and port normally
                    self.__msg_socket.bind(
                        protocol + "://" + str(address) + ":" + str(port)
                    )

                    # define socket optimizer
                    self.__msg_socket.setsockopt(zmq.LINGER, 0)

            except Exception as e:
                logger.exception(str(e))
                # otherwise raise value error if errored
                if self.__secure_mode:
                    logger.warning(
                        "Failed to activate ZMQ Security Mechanism: `{}` for this address!".format(
                            valid_security_mech[self.__secure_mode]
                        )
                    )
                if self.__multiserver_mode:
                    raise ValueError(
                        "[NetGear:ERROR] :: Multi-Server Mode, failed to connect to ports: {} with pattern: {}! Kindly recheck all parameters.".format(
                            str(port), pattern
                        )
                    )
                else:
                    raise ValueError(
                        "[NetGear:ERROR] :: Failed to bind address: {} and pattern: {}! Kindly recheck all parameters.".format(
                            (protocol + "://" + str(address) + ":" + str(port)), pattern
                        )
                    )

            # log and enable threaded queue mode
            if self.__logging:
                logger.debug("Threaded Queue Mode is enabled by default for NetGear.")
            # define deque and assign it to global var
            self.__queue = deque(maxlen=96)  # max len 96 to check overflow

            # initialize and start threading instance
            self.__thread = Thread(target=self.__update, name="NetGear", args=())
            self.__thread.daemon = True
            self.__thread.start()

            if self.__logging:
                # finally log progress
                logger.debug(
                    "Successfully Binded to address: {} with pattern: {}.".format(
                        (protocol + "://" + str(address) + ":" + str(port)), pattern
                    )
                )
                if self.__secure_mode:
                    logger.debug(
                        "Enabled ZMQ Security Mechanism: `{}` for this address, Successfully!".format(
                            valid_security_mech[self.__secure_mode]
                        )
                    )
                logger.debug("Multi-threaded Receive Mode is enabled Successfully!")
                logger.debug("Device Unique ID is {}.".format(self.__id))
                logger.debug("Receive Mode is activated successfully!")
        else:
            # otherwise default to `Send Mode`
            if address is None:
                address = "localhost"  # define address

            # check if multiserver_mode is enabled
            if self.__multiserver_mode:
                # check if unique server port address is assigned or not in multiserver_mode
                if port is None:
                    # raise error is not
                    raise ValueError(
                        "[NetGear:ERROR] :: Kindly provide a unique & valid port value at Server-end. For more information refer VidGear docs."
                    )
                else:
                    # otherwise log it
                    logger.debug(
                        "Enabling Multi-Server Mode at PORT: {} on this device!".format(
                            port
                        )
                    )
                    # assign value to global variable
                    self.port = port
            else:
                # otherwise assign local port address if None
                if port is None:
                    port = "5555"

            try:
                # initiate and handle secure mode
                if self.__secure_mode > 0:
                    # start an authenticator for this context
                    auth = ThreadAuthenticator(self.__msg_context)
                    auth.start()
                    auth.allow(str(address))  # allow current address

                    # check if `IronHouse` is activated
                    if self.__secure_mode == 2:
                        # tell authenticator to use the certificate from given valid dir
                        auth.configure_curve(
                            domain="*", location=self.__auth_publickeys_dir
                        )
                    else:
                        # otherwise tell the authenticator how to handle the CURVE requests, if `StoneHouse` is activated
                        auth.configure_curve(
                            domain="*", location=zmq.auth.CURVE_ALLOW_ANY
                        )

                # initialize and define thread-safe messaging socket
                self.__msg_socket = self.__msg_context.socket(msg_pattern[0])

                if self.__pattern == 1:
                    # if pattern is 1, define additional flags
                    self.__msg_socket.REQ_RELAXED = True
                    self.__msg_socket.REQ_CORRELATE = True
                if self.__pattern == 2 and not (self.__secure_mode):
                    self.__msg_socket.set_hwm(
                        1
                    )  # if pattern is 2, define additional optimizer

                # enable specified secure mode for the zmq socket
                if self.__secure_mode > 0:
                    # load client key
                    client_secret_file = os.path.join(
                        self.__auth_secretkeys_dir, "client.key_secret"
                    )
                    client_public, client_secret = zmq.auth.load_certificate(
                        client_secret_file
                    )
                    # load  all CURVE keys
                    self.__msg_socket.curve_secretkey = client_secret
                    self.__msg_socket.curve_publickey = client_public
                    # load server key
                    server_public_file = os.path.join(
                        self.__auth_publickeys_dir, "server.key"
                    )
                    server_public, _ = zmq.auth.load_certificate(server_public_file)
                    # inject public key to make a CURVE connection.
                    self.__msg_socket.curve_serverkey = server_public

                # connect socket to given protocol, address and port
                self.__msg_socket.connect(
                    protocol + "://" + str(address) + ":" + str(port)
                )

                # define socket options
                self.__msg_socket.setsockopt(zmq.LINGER, 0)

            except Exception as e:
                logger.exception(str(e))
                # log if errored
                if self.__secure_mode:
                    logger.warning(
                        "Failed to activate ZMQ Security Mechanism: `{}` for this address!".format(
                            valid_security_mech[self.__secure_mode]
                        )
                    )
                # raise value error
                raise ValueError(
                    "[NetGear:ERROR] :: Failed to connect address: {} and pattern: {}! Kindly recheck all parameters.".format(
                        (protocol + "://" + str(address) + ":" + str(port)), pattern
                    )
                )

            if self.__logging:
                # finally log progress
                logger.debug(
                    "Successfully connected to address: {} with pattern: {}.".format(
                        (protocol + "://" + str(address) + ":" + str(port)), pattern
                    )
                )
                if self.__secure_mode:
                    logger.debug(
                        "Enabled ZMQ Security Mechanism: `{}` for this address, Successfully!".format(
                            valid_security_mech[self.__secure_mode]
                        )
                    )
                logger.debug("This device Unique ID is {}.".format(self.__id))
                logger.debug(
                    "Send Mode is successfully activated and ready to send data!"
                )

    def __update(self):
        """
        Updates recovered frames from messaging network to the queue
        """
        # initialize frame variable
        frame = None

        # keep looping infinitely until the thread is terminated
        while not self.__exit_loop:

            # check if global termination_flag is enabled
            if self.__terminate:
                # check whether there is still frames in queue before breaking out
                if len(self.__queue) > 0:
                    continue
                else:
                    break

            # check queue buffer for overflow
            if len(self.__queue) >= 96:
                # stop iterating if overflowing occurs
                time.sleep(0.000001)
                continue

            # extract json data out of socket
            msg_json = self.__msg_socket.recv_json(flags=self.__msg_flag)

            # check if terminate_flag` received
            if msg_json["terminate_flag"]:
                # if multiserver_mode is enabled
                if self.__multiserver_mode:
                    # check and remove from which ports signal is received
                    if msg_json["port"] in self.__port_buffer:
                        # if pattern is 1, then send back server the info about termination
                        if self.__pattern == 1:
                            self.__msg_socket.send_string(
                                "Termination signal received at client!"
                            )
                        self.__port_buffer.remove(msg_json["port"])
                        if self.__logging:
                            logger.warning(
                                "Termination signal received from server at port: {}!".format(
                                    msg_json["port"]
                                )
                            )
                    # if termination signal received from all servers then exit client.
                    if not self.__port_buffer:
                        logger.warning(
                            "Termination signal received from all Servers!!!"
                        )
                        self.__terminate = True  # termination
                    continue
                else:
                    # if pattern is 1, then send back server the info about termination
                    if self.__pattern == 1:
                        self.__msg_socket.send_string(
                            "Termination signal received at client!"
                        )
                    # termination
                    self.__terminate = True
                    # notify client
                    if self.__logging:
                        logger.warning("Termination signal received from server!")
                    continue

            # extract array from socket
            msg_data = self.__msg_socket.recv(
                flags=self.__msg_flag, copy=self.__msg_copy, track=self.__msg_track
            )

            if self.__pattern != 2:
                # check if bi-directional mode is enabled
                if self.__bi_mode:
                    # handle return data
                    bi_dict = dict(data=self.__bi_data)
                    self.__msg_socket.send_json(bi_dict, self.__msg_flag)
                else:
                    # send confirmation message to server
                    self.__msg_socket.send_string(
                        "Data received on device: {} !".format(self.__id)
                    )

            # recover frame from array buffer
            frame_buffer = np.frombuffer(msg_data, dtype=msg_json["dtype"])
            # reshape frame
            frame = frame_buffer.reshape(msg_json["shape"])

            # check if encoding was enabled
            if msg_json["compression"]:
                frame = cv2.imdecode(frame, self.__compression_params)
                # check if valid frame returned
                if frame is None:
                    self.__terminate = True
                    # otherwise raise error and exit
                    raise ValueError(
                        "[NetGear:ERROR] :: `{}` Frame Decoding failed with Parameter: {}".format(
                            msg_json["compression"], self.__compression_params
                        )
                    )

            if self.__multiserver_mode:
                # check if multiserver_mode

                # save the unique port addresses
                if not msg_json["port"] in self.__port_buffer:
                    self.__port_buffer.append(msg_json["port"])

                # extract if any message from server and display it
                if msg_json["message"]:
                    self.__queue.append((msg_json["port"], msg_json["message"], frame))
                else:
                    # append recovered unique port and frame to queue
                    self.__queue.append((msg_json["port"], frame))
            else:
                # extract if any message from server if Bi-Directional Mode is enabled
                if self.__bi_mode and msg_json["message"]:
                    # append grouped frame and data to queue
                    self.__queue.append((msg_json["message"], frame))
                else:
                    # otherwise append recovered frame to queue
                    self.__queue.append(frame)

        # finally properly close the socket
        self.__msg_socket.close()

    def recv(self, return_data=None):
        """
        return the recovered frame

        :param return_data: handles return data for bi-directional mode 
        """
        # check whether `receive mode` is activated
        if not (self.__receive_mode):
            # raise value error and exit
            self.__terminate = True
            raise ValueError(
                "[NetGear:ERROR] :: `recv()` function cannot be used while receive_mode is disabled. Kindly refer vidgear docs!"
            )

        # handle bi-directional return data
        if self.__bi_mode and not (return_data is None):
            self.__bi_data = return_data

        # check whether or not termination flag is enabled
        while not self.__terminate:
            # check if queue is empty
            if len(self.__queue) > 0:
                return self.__queue.popleft()
            else:
                continue
        # otherwise return NoneType
        return None

    def send(self, frame, message=None):
        """
        send the frames over the messaging network

        :param frame(ndarray): frame array to send
        :param message(string/int): additional message for the client(s) 
        """
        # check whether `receive_mode` is disabled
        if self.__receive_mode:
            # raise value error and exit
            self.__terminate = True
            raise ValueError(
                "[NetGear:ERROR] :: `send()` function cannot be used while receive_mode is enabled. Kindly refer vidgear docs!"
            )

        # define exit_flag and assign value
        exit_flag = True if (frame is None or self.__terminate) else False

        # check whether exit_flag is False
        if not (exit_flag) and not (frame.flags["C_CONTIGUOUS"]):
            # check whether the incoming frame is contiguous
            frame = np.ascontiguousarray(frame, dtype=frame.dtype)

        # handle encoding
        if self.__compression:
            retval, frame = cv2.imencode(
                self.__compression, frame, self.__compression_params
            )
            # check if it works
            if not (retval):
                # otherwise raise error and exit
                self.__terminate = True
                raise ValueError(
                    "[NetGear:ERROR] :: Frame Encoding failed with format: {} and Parameters: {}".format(
                        self.__compression, self.__compression_params
                    )
                )

        # check if multiserver_mode is activated
        if self.__multiserver_mode:
            # prepare the exclusive json dict and assign values with unique port
            msg_dict = dict(
                terminate_flag=exit_flag,
                compression=str(self.__compression),
                port=self.port,
                pattern=str(self.__pattern),
                message=message,
                dtype=str(frame.dtype),
                shape=frame.shape,
            )
        else:
            # otherwise prepare normal json dict and assign values
            msg_dict = dict(
                terminate_flag=exit_flag,
                compression=str(self.__compression),
                message=message,
                pattern=str(self.__pattern),
                dtype=str(frame.dtype),
                shape=frame.shape,
            )

        # send the json dict
        self.__msg_socket.send_json(msg_dict, self.__msg_flag | self.__zmq.SNDMORE)
        # send the frame array with correct flags
        self.__msg_socket.send(
            frame, flags=self.__msg_flag, copy=self.__msg_copy, track=self.__msg_track
        )
        # wait for confirmation

        if self.__pattern != 2:
            # check if bi-directional data transmission is enabled
            if self.__bi_mode:
                # handle return data
                return_dict = self.__msg_socket.recv_json(flags=self.__msg_flag)
                return return_dict["data"] if return_dict else None
            else:
                # otherwise log normally
                recv_confirmation = self.__msg_socket.recv()
                # log confirmation
                if self.__logging:
                    logger.debug(recv_confirmation)

    def close(self):
        """
        Terminates the NetGear processes safely
        """
        if self.__logging:
            # log it
            logger.debug(
                "Terminating various {} Processes.".format(
                    "Receive Mode" if self.__receive_mode else "Send Mode"
                )
            )
        #  whether `receive_mode` is enabled or not
        if self.__receive_mode:
            # indicate that process should be terminated
            self.__terminate = True
            # check whether queue mode is empty
            if not (self.__queue is None):
                self.__queue.clear()
            # call immediate termination
            self.__exit_loop = True
            # wait until stream resources are released (producer thread might be still grabbing frame)
            if self.__thread is not None:
                self.__thread.join()
                self.__thread = None
                # properly handle thread exit

            # properly close the socket
            self.__msg_socket.close(linger=0)
        else:
            # indicate that process should be terminated
            self.__terminate = True
            if self.__multiserver_mode:
                # check if multiserver_mode
                # send termination flag to client with its unique port
                term_dict = dict(terminate_flag=True, port=self.port)
            else:
                # otherwise send termination flag to client
                term_dict = dict(terminate_flag=True)
            # otherwise inform client(s) that the termination has been reached
            if self.__force_close:
                # overflow socket with termination signal
                for _ in range(500):
                    self.__msg_socket.send_json(term_dict)
            else:
                self.__msg_socket.send_json(term_dict)
                # check for confirmation if available
                if self.__pattern < 2:
                    if self.__pattern == 1:
                        self.__msg_socket.recv()
            # properly close the socket
            self.__msg_socket.close(linger=0)



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from .helper import logger_handler
from ..videogear import VideoGear
from collections import deque

import cv2, inspect, platform
import asyncio, msgpack
import msgpack_numpy as m
import numpy as np
import logging as log

import zmq
import zmq.asyncio

# define logger
logger = log.getLogger("NetGear_Async")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class NetGear_Async:
    """
    NetGear_Async is NetGear API on steroids. NetGear_Async is invented for uninterrupted network performance with excellent system-resources 
    handling, insignificant latency and no fancy stuff.

    NetGear_Async is asynchronous I/O API built on AsyncIO ZmQ API and powered by state-of-the-art asyncio event loop called `uvloop`, to 
    achieve unmatchable high-speed and lag-free video streaming over the network with minimal resource constraint. 

    NetGear_Async provides complete server-client handling and options to use variable protocols/patterns but no support for any NetGear API modes.
    It supports  all four ZeroMQ messaging patterns: i.e zmq.PAIR, zmq.REQ/zmq.REP, zmq.PUB/zmq.SUB, and zmq.PUSH/zmq.PULL , whereas supported 
    protocol are: 'tcp' and 'ipc'. 


    NetGear_Async specific parameters:

    :param address(string): sets the valid network address of the server/client. Network addresses unique identifiers across the 
                            network. Its default value of this parameter is based on mode it is working, 'localhost' for Send Mode
                            and `*` for Receive Mode.

    :param port(string): sets the valid network port of the server/client. A network port is a number that identifies one side 
                            of a connection between two devices on network. It is used determine to which process or application 
                            a message should be delivered. Its default value is '5555'. 

    :param protocol(string): sets the valid messaging protocol between server and client. A network protocol is a set of established rules
                             that dictates how to format, transmit and receive data so computer network devices - from servers and 
                             routers to endpoints - can communicate regardless of the differences in their underlying infrastructures, 
                             designs or standards. As of now, supported protocol are: 'tcp', and 'ipc'. Its default value is `tcp`.

    :param pattern(int): sets the supported messaging pattern(flow of communication) between server and client. Messaging patterns are network 
                            oriented architectural pattern that describes the flow of communication between interconnecting systems.
                            vidgear provides access to ZeroMQ's pre-optimized sockets which enables you to take advantage of these patterns.
                            The supported patterns are:
                                0: zmq.PAIR -> In this, the communication is bidirectional. There is no specific state stored within the socket. 
                                                There can only be one connected peer.The server listens on a certain port and a client connects to it.
                                1. zmq.REQ/zmq.REP -> In this, ZMQ REQ sockets can connect to many servers. The requests will be
                                                        interleaved or distributed to both the servers. socket zmq.REQ will block 
                                                        on send unless it has successfully received a reply back and socket zmq.REP 
                                                        will block on recv unless it has received a request.
                                2. zmq.PUB,zmq.SUB -> Publish/Subscribe is another classic pattern where senders of messages, called publishers, 
                                                        do not program the messages to be sent directly to specific receivers, called subscribers. 
                                                        Messages are published without the knowledge of what or if any subscriber of that knowledge exists.
                                3. zmq.PUSH, zmq.PULL - > Push and Pull sockets let you distribute messages to multiple workers, arranged in a pipeline. 
                                                        A Push socket will distribute sent messages to its Pull clients evenly. This is equivalent to 
                                                        producer/consumer model but the results computed by consumer are not sent upstream but downstream 
                                                        to another pull/consumer socket.
                            Its default value is `0`(i.e zmq.PAIR).

    :param (boolean) receive_mode: set this flag to select the Netgear's Mode of operation. This basically activates `Receive Mode`(if True) and `Send Mode`(if False). 
                                    Furthermore `recv()` function will only works when this flag is enabled(i.e. `Receive Mode`) and `send()` function will only works 
                                    when this flag is disabled(i.e.`Send Mode`). Checkout VidGear docs for usage details.
                                    Its default value is False(i.e. Send Mode is activated by default).


    VideoGear Specific parameters for NetGear_Async:
    
        :param (boolean) enablePiCamera: set this flag to access PiGear or CamGear class respectively. 
                                        / This means the if enablePiCamera flag is `True`, PiGear class will be accessed 
                                        / and if `False`, the camGear Class will be accessed. Its default value is False.

        :param (boolean) stabilize: set this flag to enable access to VidGear's Stabilizer Class. This basically enables(if True) or disables(if False) 
                                        video stabilization in VidGear. Its default value is False.

    CamGear Specific supported parameters for NetGear_Async:

        :param source : take the source value for CamGear Class. Its default value is 0. Valid Inputs are:
            - Index(integer): Valid index of the video device.
            - YouTube Url(string): Youtube URL as input.
            - Network_Stream_Address(string): Incoming Stream Valid Network address. 
            - GStreamer (string) videostream Support
        :param (boolean) y_tube: enables YouTube Mode in CamGear Class, i.e If enabled the class will interpret the given source string as YouTube URL. 
                                / Its default value is False.
        :param (int) backend: set the backend of the video stream (if specified). Its default value is 0.


    PiGear Specific supported parameters for NetGear_Async:
        :param (integer) camera_num: selects the camera module index that will be used by API. 
                                /   Its default value is 0 and shouldn't be altered until unless 
                                /   if you using Raspberry Pi 3/3+ compute module in your project along with multiple camera modules. 
                                /   Furthermore, Its value can only be greater than zero, otherwise, it will throw ValueError for any negative value.
        :param (tuple) resolution: sets the resolution (width,height) in Picamera class. Its default value is (640,480).
        :param (integer) framerate: sets the framerate in Picamera class. Its default value is 25.


    Common parameters for NetGear_Async: 
        :param (string) colorspace: set colorspace of the video stream. Its default value is None.
        :param (dict) **options: parameter supported by various API (whichever being accessed).
        :param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.
        :param (integer) time_delay: sets time delay(in seconds) before start reading the frames. 
                            / This delay is essentially required for camera to warm-up. 
                            / Its default value is 0.
    """

    def __init__(
        self,
        # NetGear_Async parameters
        address=None,
        port=None,
        protocol="tcp",
        pattern=0,
        receive_mode=False,
        timeout=0.0,
        # Videogear parameters
        enablePiCamera=False,
        stabilize=False,
        source=0,
        camera_num=0,
        y_tube=False,
        backend=0,
        colorspace=None,
        resolution=(640, 480),
        framerate=25,
        time_delay=0,
        # common parameters
        logging=False,
        **options
    ):
        # enable logging if specified
        self.__logging = logging

        # define valid messaging patterns => `0`: PAIR, `1`:(REQ, REP), `2`:(SUB, PUB), `3`:(PUSH, PULL)
        valid_messaging_patterns = {
            0: (zmq.PAIR, zmq.PAIR),
            1: (zmq.REQ, zmq.REP),
            2: (zmq.PUB, zmq.SUB),
            3: (zmq.PUSH, zmq.PULL),
        }

        # check whether user-defined messaging pattern is valid
        if isinstance(pattern, int) and pattern in valid_messaging_patterns:
            # assign value
            self.__msg_pattern = pattern
            self.__pattern = valid_messaging_patterns[pattern]
        else:
            # otherwise default to 0:`zmq.PAIR`
            self.__msg_pattern = 0
            self.__pattern = valid_messaging_patterns[self.__msg_pattern]
            if self.__logging:
                logger.warning("Invalid pattern {pattern}. Defaulting to `zmq.PAIR`!".format(pattern=pattern))

        # check  whether user-defined messaging protocol is valid
        if isinstance(protocol, str) and protocol in ["tcp", "ipc"]:
            # assign value
            self.__protocol = protocol
        else:
            # else default to `tcp` protocol
            self.__protocol = "tcp"
            if self.__logging:
                logger.warning("Invalid protocol. Defaulting to `tcp`!")

        # initialize Termination flag
        self.__terminate = False
        # initialize and assign `Receive Mode`
        self.__receive_mode = receive_mode
        # initialize stream handler
        self.__stream = None
        # initialize Messaging Socket
        self.__msg_socket = None
        # initialize NetGear's configuration dictionary
        self.config = {}

        # assign timeout for Receiver end
        if timeout > 0 and isinstance(timeout, (int, float)):
            self.__timeout = float(timeout)
        else:
            self.__timeout = 10.0
        # define messaging asynchronous Context
        self.__msg_context = zmq.asyncio.Context()

        # check whether `Receive Mode` is enabled
        if receive_mode:
            # assign local ip address if None
            if address is None:
                self.__address = "*"  # define address
            else: 
                self.__address = address
            # assign default port address if None
            if port is None:
                self.__port = "5555"
            else: 
                self.__port = port
        else:
            # Handle video source if not None
            if not (source is None) and source:
                # define stream with necessary params
                self.__stream = VideoGear(
                    enablePiCamera=enablePiCamera,
                    stabilize=stabilize,
                    source=source,
                    camera_num=camera_num,
                    y_tube=y_tube,
                    backend=backend,
                    colorspace=colorspace,
                    resolution=resolution,
                    framerate=framerate,
                    logging=logging,
                    time_delay=time_delay,
                    **options
                )
                # define default frame generator in configuration
                self.config = {"generator": self.__frame_generator()}
            else:
                # else set it to None
                self.config = {"generator": None}
            # assign local ip address if None
            if address is None:
                self.__address = "localhost"
            else:
                self.__address = address
            # assign default port address if None
            if port is None:
                self.__port = "5555"
            else:
                self.__port = port
            # add server task handler
            self.task = None

        # Setup and assign uvloop event loop policy
        if platform.system() != "Windows":
            import uvloop

            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        # Retrieve event loop and assign it
        self.loop = asyncio.get_event_loop()

    def launch(self):
        """
        Launches asynchronous loop executors for various tasks
        """
        # check if receive mode enabled
        if self.__receive_mode:
            if self.__logging:
                logger.debug("Launching NetGear asynchronous generator!")
            # run loop executor for Receiver asynchronous generator
            self.loop.run_in_executor(None, self.recv_generator)
            # return instance
            return self
        else:
            # Otherwise launch Server handler
            if self.__logging:
                logger.debug("Creating NetGear asynchronous server handler!")
            # create task for Server Handler
            self.task = asyncio.ensure_future(self.__server_handler(), loop = self.loop)
            # return instance
            return self

    async def __server_handler(self):
        """
        Handles various Server-end  processes
        """
        # validate assigned frame generator in NetGear configuration
        if isinstance(self.config, dict) and "generator" in self.config:
            # check if its  assigned value is a asynchronous generator
            if self.config["generator"] is None or not inspect.isasyncgen(
                self.config["generator"]
            ):
                # otherwise raise error
                raise ValueError(
                    "[NetGear_Async:ERROR] :: Invalid configuration. Assigned generator must be a asynchronous generator function/method only!"
                )
        else:
            # raise error if validation fails
            raise RuntimeError(
                "[NetGear_Async:ERROR] :: Assigned NetGear configuration is invalid!"
            )

        # define our messaging socket
        self.__msg_socket = self.__msg_context.socket(self.__pattern[0])

        # try connecting socket to assigned protocol, address and port
        try:
            self.__msg_socket.connect(
                self.__protocol + "://" + str(self.__address) + ":" + str(self.__port)
            )
        except Exception as e:
            # log ad raise error if failed
            logger.exception(str(e))
            raise ValueError(
                "[NetGear_Async:ERROR] :: Failed to connect address: {} and pattern: {}!".format(
                    (
                        self.__protocol
                        + "://"
                        + str(self.__address)
                        + ":"
                        + str(self.__port)
                    ),
                    self.__msg_pattern,
                )
            )
        finally:
            # finally log if successful
            if self.__logging:
                logger.debug(
                    "Successfully connected to address: {} with pattern: {}.".format(
                        (
                            self.__protocol
                            + "://"
                            + str(self.__address)
                            + ":"
                            + str(self.__port)
                        ),
                        self.__msg_pattern,
                    )
                )
                logger.debug(
                    "Send Mode is successfully activated and ready to send data!"
                )
        # loop over our Asynchronous frame generator
        async for frame in self.config["generator"]:
            # check if retrieved frame is `CONTIGUOUS`
            if not (frame.flags["C_CONTIGUOUS"]):
                # otherwise make it
                frame = np.ascontiguousarray(frame, dtype=frame.dtype)
            # encode message
            msg_enc = msgpack.packb(frame, default=m.encode)
            # send it over network
            await self.__msg_socket.send_multipart([msg_enc])
            # check if bidirectional patterns
            if self.__msg_pattern < 2:
                # then receive and log confirmation
                recv_confirmation = await self.__msg_socket.recv_multipart()
                if self.__logging:
                    logger.debug(recv_confirmation)
        # send `exit` flag when done!
        await self.__msg_socket.send_multipart([b"exit"])

    async def recv_generator(self):
        """
        Asynchronous Frame Generator for NetGear's receiver-end  
        """
        # check whether `receive mode` is activated
        if not (self.__receive_mode):
            # raise Value error and exit
            self.__terminate = True
            raise ValueError(
                "[NetGear:ERROR] :: `recv_generator()` function cannot be accessed while `receive_mode` is disabled. Kindly refer vidgear docs!"
            )

        # initialize and define messaging socket
        self.__msg_socket = self.__msg_context.socket(self.__pattern[1])

        try:
            # define exclusive socket options for patterns
            if self.__msg_pattern == 2:
                self.__msg_socket.setsockopt(zmq.SUBSCRIBE, b"")
            # bind socket to the assigned protocol, address and port
            self.__msg_socket.bind(
                self.__protocol + "://" + str(self.__address) + ":" + str(self.__port)
            )
        except Exception as e:
            logger.exception(str(e))
            raise ValueError(
                "[NetGear:ERROR] :: Failed to bind address: {} and pattern: {}!".format(
                    (
                        self.__protocol
                        + "://"
                        + str(self.__address)
                        + ":"
                        + str(self.__port)
                    ),
                    self.__msg_pattern,
                )
            )
        finally:
            # finally log progress
            if self.__logging:
                logger.debug(
                    "Successfully Binded to address: {} with pattern: {}.".format(
                        (
                            self.__protocol
                            + "://"
                            + str(self.__address)
                            + ":"
                            + str(self.__port)
                        ),
                        self.__msg_pattern,
                    )
                )
                logger.debug("Receive Mode is activated successfully!")

        # loop until terminated
        while not self.__terminate:
            # get message withing timeout limit
            recvd_msg = await asyncio.wait_for(
                self.__msg_socket.recv_multipart(), timeout=self.__timeout
            )
            # check if bidirectional patterns
            if self.__msg_pattern < 2:
                # send confirmation
                await self.__msg_socket.send_multipart([b"Message Received!"])
            # terminate if exit` flag received
            if recvd_msg[0] == b"exit":
                break
            # retrieve frame from message
            frame = msgpack.unpackb(recvd_msg[0], object_hook=m.decode)
            # yield received frame
            yield frame
            # sleep for sometime
            await asyncio.sleep(0.00001)

    async def __frame_generator(self):
        """
        Default frame generator for NetGear's Server Handler 
        """
        # start stream
        self.__stream.start()
        # loop over stream until its terminated
        while not self.__terminate:
            # read frames
            frame = self.__stream.read()
            # break if NoneType
            if frame is None:
                break
            # yield frame
            yield frame
            # sleep for sometime
            await asyncio.sleep(0.00001)

    def close(self, skip_loop=False):
        """
        Terminates the NetGear Asynchronous processes safely

        :param: skip_loop(Boolean) => used if closing loop throws error
        """
        # log termination
        if self.__logging:
            logger.debug(
                "Terminating various {} Processes.".format(
                    "Receive Mode" if self.__receive_mode else "Send Mode"
                )
            )
        #  whether `receive_mode` is enabled or not
        if self.__receive_mode:
            # indicate that process should be terminated
            self.__terminate = True
        else:
            # indicate that process should be terminated
            self.__terminate = True
            # terminate stream
            if not (self.__stream is None):
                self.__stream.stop()

        # close event loop if specified
        if not (skip_loop):
            self.loop.close()



"""
===============================================
vidgear library source-code is deployed under the Apache 2.0 License:

Copyright (c) 2019 Abhishek Thakur(@abhiTronix) <abhi.una12@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
===============================================
"""

# import the necessary packages
from threading import Thread
from pkg_resources import parse_version
from .helper import capPropId
from .helper import logger_handler
from mss import mss
from mss.exception import ScreenShotError

import numpy as np
import cv2, time
import logging as log


# define logger
logger = log.getLogger("ScreenGear")
logger.addHandler(logger_handler())
logger.setLevel(log.DEBUG)


class ScreenGear:

    """
	With ScreenGear, we can easily define an area on the computer screen or an open window to record the live screen frames in 
	real-time at the expense of insignificant latency. To achieve this, ScreenGear provides a high-level multi-threaded wrapper 
	around mss python library API and also supports the flexible direct parameter manipulation. Furthermore, ScreenGear relies on 
	Threaded Queue mode for ultra-fast live frame handling and which is enabled by default.

	Threaded Queue Mode => Sequentially adds and releases frames to/from deque and handles overflow of this queue. It utilizes 
	Deques that support thread-safe, memory efficient appends and pops from either side of the deque with approximately the 
	same O(1) performance in either direction.  


	:param monitor(int): sets the Positions/Location of monitor where to grab frame from. More information can be found in the docs. 
						/ It default value is 1 (means current monitor will be used).

	:param **options(dict): can be used to pass parameters to ScreenGear Class. 
							/This attribute provides the flexibility to manipulate mss input parameters 
							/directly like the dimensions of the region of the given monitor from where the 
							frames to be grabbed. Checkout VidGear docs for usage details.

	:param (string) colorspace: set the colorspace of the video stream. Its default value is None.

	:param (boolean) logging: set this flag to enable/disable error logging essential for debugging. Its default value is False.

	"""

    def __init__(self, monitor=1, colorspace=None, logging=False, **options):

        # enable logging if specified
        self.__logging = False
        if logging:
            self.__logging = logging

        # create mss object
        self.__mss_object = mss()
        # create monitor instance for the user-defined monitor
        monitor_instance = None
        if monitor >= 0:
            try:
                monitor_instance = self.__mss_object.monitors[monitor]
            except Exception as e:
                logger.exception(str(e))
                monitor_instance = None
        else:
            raise ValueError(
                "[ScreenGear:ERROR] :: `monitor` value cannot be negative, Read Docs!"
            )

        # Initialize Queue
        self.__queue = None

        # import deque
        from collections import deque

        # define deque and assign it to global var
        self.__queue = deque(maxlen=96)  # max len 96 to check overflow
        # log it
        if logging:
            logger.debug("Enabling Threaded Queue Mode by default for ScreenGear!")

        # intiate screen dimension handler
        screen_dims = {}
        # initializing colorspace variable
        self.color_space = None

        # reformat proper mss dict and assign to screen dimension handler
        screen_dims = {
            k.strip(): v
            for k, v in options.items()
            if k.strip() in ["top", "left", "width", "height"]
        }
        # separately handle colorspace value to int conversion
        if not (colorspace is None):
            self.color_space = capPropId(colorspace.strip())
            if logging and not (self.color_space is None):
                logger.debug(
                    "Enabling `{}` colorspace for this video stream!".format(
                        colorspace.strip()
                    )
                )

        # intialize mss capture instance
        self.__mss_capture_instance = None
        try:
            # check whether user-defined dimensions are provided
            if screen_dims and len(screen_dims) == 4:
                if logging:
                    logger.debug("Setting capture dimensions: {}!".format(screen_dims))
                self.__mss_capture_instance = (
                    screen_dims  # create instance from dimensions
                )
            elif not (monitor_instance is None):
                self.__mss_capture_instance = (
                    monitor_instance  # otherwise create instance from monitor
                )
            else:
                raise RuntimeError("[ScreenGear:ERROR] :: API Failure occurred!")
            # extract global frame from instance
            self.frame = np.asanyarray(
                self.__mss_object.grab(self.__mss_capture_instance)
            )
            # intitialize and append to queue
            self.__queue.append(self.frame)
        except Exception as e:
            if isinstance(e, ScreenShotError):
                # otherwise catch and log errors
                if logging:
                    logger.exception(self.__mss_object.get_error_details())
                raise ValueError(
                    "[ScreenGear:ERROR] :: ScreenShotError caught, Wrong dimensions passed to python-mss, Kindly Refer Docs!"
                )
            else:
                raise SystemError(
                    "[ScreenGear:ERROR] :: Unable to initiate any MSS instance on this system, Are you running headless?"
                )

        # thread initialization
        self.__thread = None
        # initialize termination flag
        self.__terminate = False

    def start(self):
        """
		start the thread to read frames from the video stream
		"""
        self.__thread = Thread(target=self.__update, name="ScreenGear", args=())
        self.__thread.daemon = True
        self.__thread.start()
        return self

    def __update(self):
        """
		Update frames from stream
		"""
        # intialize frame variable
        frame = None
        # keep looping infinitely until the thread is terminated
        while not(self.__terminate):

            # check queue buffer for overflow
            if len(self.__queue) >= 96:
                # stop iterating if overflowing occurs
                time.sleep(0.000001)
                continue

            try:
                frame = np.asanyarray(
                    self.__mss_object.grab(self.__mss_capture_instance)
                )
                assert not(frame is None or np.shape(frame) == ()), "[ScreenGear:ERROR] :: Failed to retreive any valid frames!"
            except Exception as e:
                if isinstance(e, ScreenShotError):
                    raise RuntimeError(self.__mss_object.get_error_details())
                else:
                    logger.exception(str(e))
                self.__terminate = True
                continue

            if not (self.color_space is None):
                # apply colorspace to frames
                color_frame = None
                try:
                    if isinstance(self.color_space, int):
                        color_frame = cv2.cvtColor(frame, self.color_space)
                    else:
                        if self.__logging:
                            logger.warning(
                                "Global color_space parameter value `{}` is not a valid!".format(
                                    self.color_space
                                )
                            )
                        self.color_space = None
                except Exception as e:
                    # Catch if any error occurred
                    self.color_space = None
                    if self.__logging:
                        logger.exception(str(e))
                        logger.warning("Input colorspace is not a valid colorspace!")
                if not (color_frame is None):
                    self.frame = color_frame
                else:
                    self.frame = frame
            else:
                self.frame = frame
            # append to queue
            self.__queue.append(self.frame)
        # finally release mss resources
        self.__mss_object.close()

    def read(self):
        """
		return the frame
		"""
        # check whether or not termination flag is enabled
        while not self.__terminate:
            # check if queue is empty
            if len(self.__queue) > 0:
                return self.__queue.popleft()
            else:
                continue
        # otherwise return NoneType
        return None

    def stop(self):
        """
		Terminates the Read process
		"""
        if self.__logging:
            logger.debug("Terminating ScreenGear Processes.")
        # indicate that the thread should be terminated
        self.__terminate = True
        # terminate Threaded queue mode seperately
        if not (self.__queue is None):
            self.__queue.clear()
        # wait until stream resources are released (producer thread might be still grabbing frame)
        if self.__thread is not None:
            self.__thread.join()
            # properly handle thread exit
